import uuid

from django.test import Client, TestCase, override_settings
from django.urls import reverse

from core.models import Conference, ConferenceMember, PlatformUser

TEST_CONF_ID = uuid.uuid4()


@override_settings(SELECTED_CONFERENCE_ID=TEST_CONF_ID)
class EngelsystemTestCase(TestCase):
    dummy_pswd = 'Spagh3ttiMon$ter'

    def setUp(self):
        self.conference1 = Conference(id=TEST_CONF_ID, slug='foo', name='Foo Conference', is_public=True)
        self.conference1.save()

        self.human_user = PlatformUser(username='bernd', user_type=PlatformUser.Type.HUMAN)
        self.human_user.save()
        self.human_cm = ConferenceMember(conference=self.conference1, user=self.human_user)
        self.human_cm.save()

        self.angel_url = reverse('api:user-angel', kwargs={'username': 'bernd'})

        self.service_user = PlatformUser(username='engelsystem', user_type=PlatformUser.Type.SERVICE)
        self.service_user.set_password(self.dummy_pswd)
        self.service_user.save()
        ConferenceMember(conference=self.conference1, user=self.service_user, roles=['engelsystem']).save()

    def testUnauthorized(self):
        c = Client()

        response = c.get(self.angel_url)
        self.assertEqual(401, response.status_code, f'Unauthenticated user got unexpected response: {response.content}')

        c.force_login(self.human_user)
        response = c.get(self.angel_url)
        self.assertEqual(401, response.status_code, f'Human user got unexpected response: {response.content}')

    def get_auth_token(self, client, username, password):
        # do token-based authentication
        response = client.post('/api/auth/get-token', {'username': username, 'password': password})
        self.assertEqual(response.status_code, 200, f'Expected token auth endpoint return HTTP 200 instead of {response.status_code}.')
        response_data = response.json()
        self.assertTrue('token' in response_data)

        return response_data['token']

    def testFetch(self):
        # configure client
        c = Client()
        auth_token = self.get_auth_token(c, self.service_user.username, self.dummy_pswd)

        # fetch the current status
        response = c.get(self.angel_url, HTTP_AUTHORIZATION=f'Token {auth_token}')
        self.assertEqual(response.status_code, 200, f'Unexpected endpoint response: {response.content}.')
        endpoint_data = response.json()

        # verify all expected fields are there
        for fld in ['active_angel', 'angel_types']:
            self.assertTrue(fld in endpoint_data, f'data is missing "{fld}"')

    def testUpdate(self):
        # configure client
        c = Client()
        auth_token = self.get_auth_token(c, self.service_user.username, self.dummy_pswd)

        # sanity check: human should not be an angel and without roles
        response = c.get(self.angel_url, HTTP_AUTHORIZATION=f'Token {auth_token}')
        self.assertEqual(200, response.status_code, f'Unexpected endpoint response: {response.content}')
        data = response.json()
        self.assertTrue(not data['active_angel'])
        self.assertEqual(0, len(data['angel_types'] or []))

        # send backend_status update
        data = {'active_angel': True}
        response = c.post(self.angel_url, content_type='application/json', data=data, HTTP_AUTHORIZATION=f'Token {auth_token}')
        self.assertEqual(response.status_code, 200, f'unexpected response when POSTting active_angel flag: {response.content}')

        # verify that the update is in the database
        self.human_cm.refresh_from_db()
        self.assertTrue(self.human_cm.active_angel)

        # now update two fields: backend_status and backend_data
        data = {'active_angel': False, 'angel_types': ['foo', 'bar']}
        self.assertEqual(200, c.post(self.angel_url, content_type='application/json', data=data, HTTP_AUTHORIZATION=f'Token {auth_token}').status_code)

        # verify both have been set
        self.human_cm.refresh_from_db()
        self.assertFalse(self.human_cm.active_angel)
        self.assertTrue('foo' in self.human_cm.roles and 'bar' in self.human_cm.roles)
