import logging

from rest_framework.decorators import api_view
from rest_framework.exceptions import NotFound
from rest_framework.generics import ListCreateAPIView
from rest_framework.response import Response

from django.http import HttpResponse
from django.shortcuts import get_object_or_404

from core.models.badges import Badge, BadgeToken
from core.models.conference import Conference
from core.models.users import PlatformUser

from api.permissions import HasIssuingToken, IsAssemblyManager, IsSuperUser
from api.serializers import BadgeSerializer, BadgeTokenSerializer
from api.views.mixins import ConferenceSlugAssemblyMixin

logger = logging.getLogger(__name__)


class BadgeListCreate(ListCreateAPIView):
    model = Badge
    serializer_class = BadgeSerializer

    def get_queryset(self):
        queryset = self.model.objects.all()
        assembly = self.kwargs.get('assembly')
        return queryset.filter(issuing_assembly__slug=assembly)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context.update({'assembly': self.kwargs.get('assembly'), 'conference': self.kwargs.get('conference')})
        return context


class BadgeTokenListCreate(ConferenceSlugAssemblyMixin, ListCreateAPIView):
    model = BadgeToken
    serializer_class = BadgeTokenSerializer
    permission_classes = [IsAssemblyManager | IsSuperUser | HasIssuingToken]

    def get_queryset(self):
        queryset = BadgeToken.objects.all()
        assembly = self.kwargs.get('assembly')
        badge = self.kwargs.get('pk')
        return queryset.filter(badge=badge, badge__issuing_assembly__slug=assembly)

    def get_serializer_context(self):
        if self.kwargs.get('pk'):
            badge = self.kwargs.get('pk')
        else:
            try:
                badge = Badge.objects.get(issuing_token=self.kwargs.get('issuing_token'), conference=self.conference).pk
            except Badge.DoesNotExist:
                raise NotFound('Badge matching issuing_token not found')
        context = super().get_serializer_context()
        context.update({'badge': badge})
        return context


@api_view(['POST'])
def redeem_badge_token(request, conference, **kwargs):
    redeem_token = request.POST.get('token', request.data.get('token', None))
    if redeem_token is None:
        return HttpResponse(status=400)

    username = request.POST.get('username', request.data.get('username', None))
    user = get_object_or_404(PlatformUser, username=username, is_active=True)

    # TODO: Proper Access Checking
    conf = get_object_or_404(Conference, slug=conference)
    if not user.is_in_conference(conf):
        return HttpResponse(status=404)

    badge_token = get_object_or_404(BadgeToken, token=redeem_token.strip())
    if badge_token.badge.conference.slug != conference:
        return HttpResponse(status=400)

    badge_token.redeem(user, False)

    return Response({'badge': f'{badge_token.badge}', 'user': f'{user.username}'})
