/**
 * Sets up dynamic addition of form rows.
 *
 * Prerequirements:
 * - Given there is a Django form with the ID FORM_ID, e.g. "links"
 * - The form container element must have an ID like FORM_ID-form, e.g. "links-form"
 * - There must be an add button with an ID like FORM_ID-add, e.g. "links-add"
 *
 * @param {string} formName - Form name, e.g. 'links'. Used to find the form elements:
 *                            Django's total form elements: id_${formName}-TOTAL_FORMS
 *                            The form itself: #${formName}-form
 *                            The add button: #${formName}-add
 * @param {string} formRowSelector - The form rows selector. Searched in the form found by #${formName}-form.
 */
const setUpDynamicFormAdd = function (formName, formRowSelector) {
  const totalFormsElementSelector = `#id_${formName}-TOTAL_FORMS`;
  const totalFormsElement = document.querySelector(totalFormsElementSelector);

  if (totalFormsElement === null) {
    console.warn(`[setUpDynamicFormAdd] Total forms element not found (selector '${totalFormsElementSelector}')`);
    return;
  }

  const containerSelector = `#${formName}-form`;
  const container = document.querySelector(containerSelector);

  if (totalFormsElement === null) {
    console.warn(`[setUpDynamicFormAdd] Container element not found (selector '${containerSelector}')`);
    return;
  }

  const addButtonSelector = `#${formName}-add`;
  const addButton = document.querySelector(addButtonSelector);

  if (addButton === null) {
    console.warn(`[setUpDynamicFormAdd] Add button element not found (selector '${addButtonSelector}')`);
    return;
  }

  const rows = container.querySelectorAll(formRowSelector);
  const lastRow = rows[rows.length - 1];

  if (lastRow === null || lastRow === undefined) {
    console.warn(`[setUpDynamicFormAdd] Last form row not found (selector '${formRowSelector}')`);
    return;
  }

  // Last row should be an empty one, that can be used as template. Copy on setup.
  const templateRow = lastRow.cloneNode(true);

  let formNum = rows.length - 1;

  /**
   * @param {MouseEvent} event
   */
  function handleAddItem(event) {
    event.preventDefault();

    // Copy the template and replace the form row numbers
    formNum++;
    const newRow = templateRow.cloneNode(true);
    const idRegex = RegExp(`${formName}-(\\d){1}-`, "g");
    newRow.innerHTML = newRow.innerHTML.replace(idRegex, `${formName}-${formNum}-`);

    // Add the new form row to the container
    container.appendChild(newRow);

    // Increment total rows by one, so that Django is aware of it
    totalFormsElement.setAttribute("value", `${formNum + 1}`);
  }

  addButton.addEventListener("click", handleAddItem);
};
