from uuid import uuid4

from django.urls import reverse

from core.models import (
    Assembly,
    AssemblyMember,
    ConferenceMember,
    Invitation,
    PlatformUser,
    Team,
    TeamMember,
)

from backoffice.tests.base import BackOfficeTestCase
from backoffice.tests.invitations.habitat import HabitatInvitationViewsTestCase
from backoffice.tests.invitations.teams import MemberToTeamInvitationViewsTestCase, TeamToUserInvitationViewsTestCase


class InvitationTestCase(BackOfficeTestCase):
    def test_invalid_invitation_type(self):
        self.client.force_login(self.staff)
        response = self.client.post(
            reverse(
                'backoffice:invitation-send',
                kwargs={'type': 'invalid', 'requester_id': uuid4()},
            ),
        )
        self.assertContains(response, 'Bad Request', status_code=400)


class InvitationListViewTestCase(BackOfficeTestCase):
    def setUp(self):
        super().setUp()

        self.assembly = Assembly.objects.create(name='Assembly 1', slug='assembly_1', conference=self.conf)
        self.assembly_manager_am = AssemblyMember.objects.create(
            assembly=self.assembly,
            member=self.user,
            is_representative=True,
            can_manage_assembly=True,
        )
        self.habitat = Assembly.objects.create(name='Habitat 1', slug='habitat_1', conference=self.conf, hierarchy=Assembly.Hierarchy.CLUSTER)
        self.habitat_manager = PlatformUser.objects.create(username='habitat_manager')
        self.habitat_manager_cm = ConferenceMember.objects.create(conference=self.conf, user=self.habitat_manager)
        self.habitat_manager_am = AssemblyMember.objects.create(
            assembly=self.habitat,
            member=self.habitat_manager,
            is_representative=True,
            can_manage_assembly=True,
        )

        self.team = Team.objects.create(name='team 1', conference=self.conf)
        self.team_manager = PlatformUser.objects.create(username='assembly_manager')
        self.team_manager_cm = ConferenceMember.objects.create(conference=self.conf, user=self.team_manager)
        self.team_manager_am = TeamMember.objects.create(
            team=self.team,
            user=self.team_manager,
            can_manage=True,
        )

        Invitation.objects.create(
            requester=self.user,
            requested=self.team,
            conference=self.conf,
            state=Invitation.RequestsState.REQUESTED,
            type=Invitation.InvitationType.MEMBER_TO_TEAM,
        )

        Invitation.objects.create(
            requester=self.user,
            requested=self.team,
            conference=self.conf,
            state=Invitation.RequestsState.WITHDRAWN,
            type=Invitation.InvitationType.MEMBER_TO_TEAM,
        )
        Invitation.objects.create(
            requester=self.user,
            requested=self.team,
            conference=self.conf,
            state=Invitation.RequestsState.REJECTED,
            type=Invitation.InvitationType.MEMBER_TO_TEAM,
        )
        Invitation.objects.create(
            requester=self.habitat,
            requested=self.assembly,
            conference=self.conf,
            state=Invitation.RequestsState.REQUESTED,
            type=Invitation.InvitationType.HABITAT,
        )
        Invitation.objects.create(
            requester=self.assembly,
            requested=self.habitat,
            conference=self.conf,
            state=Invitation.RequestsState.REQUESTED,
            type=Invitation.InvitationType.HABITAT,
        )

    def test_invitation_list(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('backoffice:invitations'))
        self.assertContains(response, 'team 1')
        self.assertContains(response, 'Habitat 1', count=2)
        self.assertContains(response, 'Assembly 1', count=2)
        self.assertEqual(len(response.context['invitations']), 3)


__all__ = [
    'HabitatInvitationViewsTestCase',
    'InvitationListViewTestCase',
    'InvitationTestCase',
    'MemberToTeamInvitationViewsTestCase',
    'TeamToUserInvitationViewsTestCase',
]
