from unittest.mock import patch

from django.urls import reverse
from django.utils.translation import activate
from django.utils.translation import gettext as _

from core.models import (
    ConferenceMember,
    PlatformUser,
    Team,
    TeamMember,
)
from core.tests.mock import mocktrans

from backoffice.tests.base import BackOfficeTestCase


class TeamListViewTestCase(BackOfficeTestCase):
    def setUp(self):
        super().setUp()
        self.teams = {}
        for team in ['team1', 'team2', 'team3']:
            self.teams[team] = Team.objects.create(
                name=team,
                conference=self.conf,
            )
        self.teams['team1'].require_staff = True
        TeamMember.objects.create(team=self.teams['team1'], user=self.staff)
        TeamMember.objects.create(team=self.teams['team1'], user=self.admin, can_manage=True)
        TeamMember.objects.create(team=self.teams['team2'], user=self.staff, can_manage=True)
        TeamMember.objects.create(team=self.teams['team3'], user=self.admin)

    def test_team_list_unauthenticated(self):
        activate('en')
        response = self.client.get(reverse('backoffice:teams'))
        self.assertRedirects(response, reverse('backoffice:login') + '?next=' + reverse('backoffice:teams'))

    def test_team_list_admin(self):
        self.client.force_login(self.admin)
        with patch('backoffice.views.teams.teams._', mocktrans):
            response = self.client.get(reverse('backoffice:teams'))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/teams/list.html')
        self.assertQuerySetEqual(Team.objects.all().order_by('name'), response.context['teams'])
        for team in self.teams:
            self.assertContains(response, team)
        self.assertContains(response, _('Team__create__button'))

    def test_team_list_staff(self):
        self.client.force_login(self.staff)
        response = self.client.get(reverse('backoffice:teams'))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/teams/list.html')
        self.assertQuerySetEqual(Team.objects.all().order_by('name'), response.context['teams'])
        for team in self.teams:
            self.assertContains(response, team)
        self.assertNotContains(response, _('Team__create__button'))

    def test_team_list_user(self):
        self.client.force_login(self.user)
        with patch('backoffice.views.teams.teams._', mocktrans):
            response = self.client.get(reverse('backoffice:teams'))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/teams/list.html')
        self.assertQuerySetEqual(Team.objects.filter(require_staff=False).order_by('name'), response.context['teams'])
        for team in self.teams:
            self.assertContains(response, team)
        self.assertNotContains(response, _('Team__create__button'))


class TeamDetailViewTestCase(BackOfficeTestCase):
    def setUp(self):
        super().setUp()
        self.user_2 = PlatformUser.objects.create_user(
            username='user2',
        )
        self.user_2_cm = ConferenceMember.objects.create(conference=self.conf, user=self.user_2)
        self.team = Team.objects.create(
            name='team',
            conference=self.conf,
        )
        self.team_member = TeamMember.objects.create(team=self.team, user=self.admin)
        self.team_member2 = TeamMember.objects.create(team=self.team, user=self.staff, can_manage=True)
        self.team_member3 = TeamMember.objects.create(team=self.team, user=self.user)

    def test_team_detail_unauthenicated(self):
        activate('en')
        response = self.client.get(reverse('backoffice:team', kwargs={'uuid': self.team.uuid}))
        self.assertRedirects(response, reverse('backoffice:login') + '?next=' + reverse('backoffice:team', kwargs={'uuid': self.team.uuid}))

    def test_team_detail_admin(self):
        self.client.force_login(self.admin)
        response = self.client.get(reverse('backoffice:team', kwargs={'uuid': self.team.uuid}))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/teams/detail.html')
        self.assertEqual(self.team, response.context['team'])
        self.assertContains(response, _('Team__edit__button'))
        self.assertContains(response, _('Team__delete__button'))
        self.assertNotContains(response, _('TeamMember__join'))

    def test_team_detail_staff(self):
        self.client.force_login(self.staff)
        response = self.client.get(reverse('backoffice:team', kwargs={'uuid': self.team.uuid}))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/teams/detail.html')
        self.assertEqual(self.team, response.context['team'])
        self.assertContains(response, _('Team__edit__button'))
        self.assertNotContains(response, _('Team__delete__button'))
        self.assertNotContains(response, _('TeamMember__join'))

    def test_team_detail_team_member(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('backoffice:team', kwargs={'uuid': self.team.uuid}))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/teams/detail.html')
        self.assertEqual(self.team, response.context['team'])
        self.assertNotContains(response, _('Team__edit__button'))
        self.assertNotContains(response, _('Team__delete__button'))
        self.assertNotContains(response, _('TeamMember__join'))

    def test_team_detail_non_team_member(self):
        self.client.force_login(self.user_2)
        response = self.client.get(reverse('backoffice:team', kwargs={'uuid': self.team.uuid}))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/teams/detail.html')
        self.assertEqual(self.team, response.context['team'])
        self.assertNotContains(response, _('Team__edit__button'))
        self.assertNotContains(response, _('Team__delete__button'))
        self.assertContains(response, _('TeamMember__join'))


class TeamCreateViewTestCase(BackOfficeTestCase):
    def test_team_create_unauthenicated(self):
        activate('en')
        response = self.client.get(reverse('backoffice:team-create'))
        self.assertRedirects(response, reverse('backoffice:login') + '?next=' + reverse('backoffice:team-create'))

    def test_team_create_admin(self):
        self.client.force_login(self.admin)
        response = self.client.get(reverse('backoffice:team-create'))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/teams/create_edit.html')
        self.assertContains(response, _('create'))
        self.assertContains(response, _('Team__create__title'))

        response = self.client.post(reverse('backoffice:team-create'), data={'name': 'team'})
        self.assertTrue(Team.objects.filter(name='team').exists())
        team = Team.objects.get(name='team')
        self.assertRedirects(response, reverse('backoffice:team', kwargs={'uuid': team.uuid}))

    def test_team_create_staff(self):
        self.client.force_login(self.staff)
        response = self.client.get(reverse('backoffice:team-create'))
        self.assertEqual(response.status_code, 403)

    def test_team_create_team_member(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('backoffice:team-create'))
        self.assertEqual(response.status_code, 403)


class TeamUpdateViewTestCase(BackOfficeTestCase):
    def setUp(self):
        super().setUp()
        self.user_2 = PlatformUser.objects.create_user(
            username='user2',
        )
        self.user_2_cm = ConferenceMember.objects.create(conference=self.conf, user=self.user_2)
        self.team = Team.objects.create(
            name='team',
            conference=self.conf,
        )
        self.team_member = TeamMember.objects.create(team=self.team, user=self.admin)
        self.team_member2 = TeamMember.objects.create(team=self.team, user=self.staff, can_manage=True)
        self.team_member3 = TeamMember.objects.create(team=self.team, user=self.user)

    def test_team_update_unauthenicated(self):
        activate('en')
        response = self.client.get(reverse('backoffice:team-edit', kwargs={'uuid': self.team.uuid}))
        self.assertRedirects(response, reverse('backoffice:login') + '?next=' + reverse('backoffice:team-edit', kwargs={'uuid': self.team.uuid}))

    def test_team_edit_admin(self):
        self.client.force_login(self.admin)
        response = self.client.get(reverse('backoffice:team-edit', kwargs={'uuid': self.team.uuid}))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/teams/create_edit.html')
        self.assertNotContains(response, _('create'))
        # TODO: Add test for title
        self.assertContains(response, _('Team__edit__submit'))

        response = self.client.post(reverse('backoffice:team-edit', kwargs={'uuid': self.team.uuid}), data={'name': 'team'})
        self.assertTrue(Team.objects.filter(name='team').exists())
        team = Team.objects.get(name='team')
        self.assertRedirects(response, reverse('backoffice:team', kwargs={'uuid': team.uuid}))

    def test_team_edit_staff(self):
        self.client.force_login(self.staff)
        response = self.client.get(reverse('backoffice:team-edit', kwargs={'uuid': self.team.uuid}))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/teams/create_edit.html')
        self.assertNotContains(response, _('create'))
        # TODO: Add test for title
        self.assertContains(response, _('Team__edit__submit'))

        response = self.client.post(
            reverse('backoffice:team-edit', kwargs={'uuid': self.team.uuid}),
            data={'name': 'team', 'description_de': 'description_de', 'description_en': 'description_en'},
        )
        self.assertTrue(Team.objects.filter(name='team').exists())
        self.assertRedirects(response, reverse('backoffice:team', kwargs={'uuid': self.team.uuid}))
        self.team.refresh_from_db()
        self.assertEqual(self.team.description_de, 'description_de')
        self.assertEqual(self.team.description_en, 'description_en')

    def test_team_edit_team_member(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('backoffice:team-edit', kwargs={'uuid': self.team.uuid}))
        self.assertEqual(response.status_code, 403)


class TeamDeleteViewTestCase(BackOfficeTestCase):
    def setUp(self):
        super().setUp()
        self.team = Team.objects.create(
            name='team',
            conference=self.conf,
        )
        self.team_member = TeamMember.objects.create(team=self.team, user=self.admin)
        self.team_member2 = TeamMember.objects.create(team=self.team, user=self.staff, can_manage=True)
        self.team_member3 = TeamMember.objects.create(team=self.team, user=self.user)

    def test_team_delete_unauthenicated(self):
        activate('en')
        response = self.client.get(reverse('backoffice:team-delete', kwargs={'uuid': self.team.uuid}))
        self.assertRedirects(response, reverse('backoffice:login') + '?next=' + reverse('backoffice:team-delete', kwargs={'uuid': self.team.uuid}))

    def test_team_delete_admin(self):
        self.client.force_login(self.admin)
        response = self.client.get(reverse('backoffice:team-delete', kwargs={'uuid': self.team.uuid}))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/components/confirmation_modal.html')
        self.assertContains(response, _('Team__delete__warning__header'))
        self.assertContains(response, _('Team__delete__submit'))

        response = self.client.post(reverse('backoffice:team-delete', kwargs={'uuid': self.team.uuid}))
        # No confirmation, should not delete
        self.assertTrue(Team.objects.filter(name='team').exists())
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'backoffice/components/confirmation_modal.html')
        self.assertContains(response, _('Team__delete__warning__header'))
        self.assertContains(response, _('Team__delete__submit'))

        response = self.client.post(reverse('backoffice:team-delete', kwargs={'uuid': self.team.uuid}), data={'confirmation': 'true'})
        self.assertFalse(Team.objects.filter(name='team').exists())
        self.assertRedirects(response, reverse('backoffice:teams'))

    def test_team_delete_staff(self):
        self.client.force_login(self.staff)
        response = self.client.get(reverse('backoffice:team-delete', kwargs={'uuid': self.team.uuid}))
        self.assertEqual(response.status_code, 403)

    def test_team_delete_team_member(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('backoffice:team-delete', kwargs={'uuid': self.team.uuid}))
        self.assertEqual(response.status_code, 403)
