import logging

from django.contrib import messages
from django.db.models import QuerySet
from django.http import Http404
from django.shortcuts import redirect
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView
from django.views.generic.edit import FormView, UpdateView

from core.models import ActivityLogChange, ActivityLogEntry
from core.models.assemblies import AssemblyMember
from core.models.users import PlatformUser

from backoffice.forms import (
    AssemblyMemberAddForm,
    AssemblyMemberEditForm,
)
from backoffice.views.mixins import AssemblyMixin

logger = logging.getLogger(__name__)


class MemberListView(AssemblyMixin, ListView):
    template_name = 'backoffice/assembly_members.html'
    model = AssemblyMember

    assembly_url_param = 'pk'
    assembly_management = True

    def get_queryset(self):
        return AssemblyMember.objects.manageable_by_user_for_assembly(user=self.request.user, assembly=self.assembly).prefetch_related(
            'member__communication_channels'
        )

    def get_context_data(self, **kwargs):
        return {
            **super().get_context_data(**kwargs),
            'form': AssemblyMemberAddForm(),
        }

    def post(self, *args, **kwargs):
        for data_pair in self.request.POST:
            if '-' not in data_pair:
                continue

            k, v = data_pair.split('-')

            if k == 'hide':
                m = self.get_queryset().select_related('member').get(member_id=int(v))
                m.show_public = False
                m.save()
                logger.info(
                    'hide "%(member)s" on assembly %(assembly)s by %(user)s', {'member': m.member, 'assembly': self.assembly, 'user': self.request.user}
                )
                messages.success(self.request, format_lazy(_('Assembly__members__hidden'), user=m.member))
                self.assembly.log_activity(
                    user=self.request.user,
                    kind=ActivityLogEntry.Kind.ENTITY if not self.staff_access else ActivityLogEntry.Kind.STAFF,
                    **{
                        'members.modified': m.member.username,
                        f'members[{m.member.username}].show_public': ActivityLogChange(
                            old='yes',
                            new='no',
                        ),
                    },
                )

            elif k == 'show':
                m = self.get_queryset().select_related('member', 'assembly').get(member_id=int(v))
                # TODO: Update after deciding oh one or more conferences in #648
                if not m.member.conferences.filter(conference_id=m.assembly.conference_id).exists():
                    messages.error(self.request, format_lazy(_('Assembly__members__no_member_cant_display'), user=m.member))
                    continue
                m.show_public = True
                m.save()
                logger.info(
                    'show "%(member)s" on assembly %(assembly)s by %(user)s', {'member': m.member, 'assembly': self.assembly, 'user': self.request.user}
                )
                messages.success(self.request, format_lazy(_('Assembly__members__shown'), user=m.member))
                self.assembly.log_activity(
                    user=self.request.user,
                    kind=ActivityLogEntry.Kind.ENTITY if not self.staff_access else ActivityLogEntry.Kind.STAFF,
                    **{
                        'members.modified': m.member.username,
                        f'members[{m.member.username}].show_public': ActivityLogChange(
                            old='no',
                            new='yes',
                        ),
                    },
                )

            elif k == 'delete':
                if int(v) == self.request.user.id and not self.staff_access:
                    messages.error(self.request, format_lazy(_('Assembly__members__cant_remove_self')))
                    continue
                m = self.get_queryset().select_related('member').get(member_id=int(v))
                m.delete()
                logger.info(
                    'removed "%(member)s" from assembly %(assembly)s by %(user)s', {'member': m.member, 'assembly': self.assembly, 'user': self.request.user}
                )
                messages.success(self.request, format_lazy(_('Assembly__members__removed'), user=m.member))
                self.assembly.log_activity(
                    user=self.request.user,
                    kind=ActivityLogEntry.Kind.ENTITY if not self.staff_access else ActivityLogEntry.Kind.STAFF,
                    **{
                        'members.removed': ActivityLogChange(
                            old=m.member.username,
                        ),
                    },
                )

        return redirect('backoffice:assembly-members', pk=self.assembly.pk)


class MemberCreateView(AssemblyMixin, FormView):
    form_class = AssemblyMemberAddForm
    assembly_management = True

    def form_valid(self, form):
        data = form.cleaned_data
        try:
            u = PlatformUser.objects.only('id', 'username').get(is_active=True, username=data['name'])

            m, created = self.assembly.members.get_or_create(member=u, defaults={'show_public': False})

            if created:
                m.show_public = False
                m.save()
                messages.success(self.request, format_lazy(_('Assembly__members__added'), user=str(u)))
                logger.info('added "%(member)s" to assembly %(assembly)s by %(user)s', {'member': u, 'assembly': self.assembly, 'user': self.request.user})
                self.assembly.log_activity(
                    user=self.request.user,
                    kind=ActivityLogEntry.Kind.ENTITY if not self.staff_access else ActivityLogEntry.Kind.STAFF,
                    **{
                        'members.added': ActivityLogChange(
                            new=m.member.username,
                        ),
                    },
                )

            else:  # noqa: PLR5501
                if m.member == self.request.user and not self.staff_access:
                    messages.error(self.request, format_lazy(_('Assembly__members__cant_modify_self')))
                else:
                    messages.info(self.request, format_lazy(_('Assembly__members__exist'), user=str(u)))

            return redirect('backoffice:assembly-members', pk=self.assembly.pk)

        except PlatformUser.DoesNotExist:
            messages.error(self.request, _('Unknown User') + ': ' + data['name'])
            return redirect('backoffice:assembly-members', pk=self.assembly.pk)


class MembersUpdateView(AssemblyMixin, UpdateView):
    assembly_management = True
    model = AssemblyMember
    template_name = 'backoffice/assembly_members_edit.html'
    form_class = AssemblyMemberEditForm

    def get_object(self, queryset: QuerySet['AssemblyMember'] | None = None):
        queryset = queryset or self.assembly.members.all()
        try:
            return queryset.get(member__username=self.request.resolver_match.kwargs.get('uname'))
        except AssemblyMember.DoesNotExist as e:
            raise Http404(_('No %(verbose_name)s found matching the query') % {'verbose_name': AssemblyMember._meta.verbose_name}) from e

    def form_valid(self, form):
        data = form.instance
        data.save()
        u = data.member.username
        messages.success(self.request, '* ' + str(u))
        logger.info('modified member "%(member)s" in assembly %(assembly)s by %(user)s', {'member': u, 'assembly': self.assembly, 'user': self.request.user})

        changes: dict[str, str | ActivityLogChange] = {
            'members.changed': f'*{u}',
        }
        for fld_name, field in form.fields.items():
            old_value = form.get_initial_for_field(field, fld_name)
            new_value = getattr(data, fld_name)
            if old_value != new_value:
                changes[f'members[{u}].{fld_name}'] = ActivityLogChange(old=str(old_value), new=str(new_value))
        self.assembly.log_activity(
            user=self.request.user,
            kind=ActivityLogEntry.Kind.ENTITY if not self.staff_access else ActivityLogEntry.Kind.STAFF,
            **changes,
        )
        return redirect('backoffice:assembly-members', pk=self.assembly.pk)
