import logging

from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import View
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import CreateView, UpdateView

from core.models.badges import Badge, BadgeToken

from backoffice.views.badges.mixins import BadgeTokenFormMixin
from backoffice.views.mixins import AssemblyMixin

logger = logging.getLogger(__name__)


class BadgeRenewIssuingTokenView(AssemblyMixin, View):
    assembly_url_param = 'assembly'
    assembly_management = True

    def post(self, *args, **kwargs):
        badge = Badge.objects.get(
            conference=self.conference,
            issuing_assembly=self.assembly,
            pk=kwargs.get('pk'),
        )

        badge.reset_token()
        logger.info(
            'Issuing token for Badge "%(badge_name)s" in assembly %(assembly)s reset by %(user)s',
            {'badge_name': badge.name, 'assembly': self.assembly, 'user': self.request.user},
        )

        ctx = self.get_context_data()
        ctx.update({'badge': badge})
        return render(self.request, 'backoffice/assembly_badge_renewed.html', ctx)


class BadgeTokenCreateView(BadgeTokenFormMixin, AssemblyMixin, CreateView):
    def get_form_kwargs(self):
        return {'create': True, **super().get_form_kwargs()}

    def get_context_data(self, **kwargs):
        return {
            **super().get_context_data(**kwargs),
            'new': True,
        }

    def get_success_url(self):
        messages.success(self.request, _('BadgeToken__create__success %(token)s') % {'token': self.object.token})
        return super().get_success_url()

    def form_invalid(self, form, formset):
        messages.error(self.request, _('BadgeToken__create__failure'))
        return super().form_invalid(form, formset)


class BadgeTokenUpdateView(BadgeTokenFormMixin, AssemblyMixin, UpdateView):
    def get_success_url(self):
        messages.success(self.request, _('BadgeToken__edit__success %(token)s') % {'token': self.object.token})
        return super().get_success_url()

    def form_invalid(self, form, formset):
        messages.error(self.request, _('BadgeToken__edit__failure %(token)s') % {'token': self.object.token})
        return super().form_invalid(form, formset)


class BadgeTokenToggleActiveView(AssemblyMixin, SingleObjectMixin, View):
    assembly_url_param = 'assembly'
    assembly_management = True
    model = BadgeToken

    def post(self, *args, **kwargs):
        token = self.get_object()
        if not BadgeToken.type_is(token):
            raise ValueError('Object is not a BadgeToken')
        token.active = not token.active
        token.save()
        messages.success(
            self.request,
            _('BadgeToken__activated %(token)s') % {'token': token} if token.active else _('BadgeToken__deactivated %(token)s') % {'token': token},
        )
        logger.info(
            'BadgeToken "%(badge_token)s" in %(assembly)s %(status)s by %(user)s',
            {'badge_name': token, 'status': 'activated' if token.active else 'deactivated', 'assembly': self.assembly, 'user': self.request.user},
        )
        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self, *args, **kwargs):
        return reverse('backoffice:assembly-badge', kwargs={'assembly': self.assembly.pk, 'pk': self.kwargs['badge']})
