import logging

from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import TemplateView
from django.views.generic.edit import CreateView, DeleteView, UpdateView

from core.models.conference import ConferenceExportCache
from core.models.events import Event
from core.models.projects import Project

from backoffice.views.events.mixins import EventPublicationMixin, SoSFormMixin
from backoffice.views.mixins import ConferenceLoginRequiredMixin

logger = logging.getLogger(__name__)


class SelfOrganizedContentListView(ConferenceLoginRequiredMixin, TemplateView):
    template_name = 'backoffice/sos.html'

    def get_context_data(self, *args, **kwargs):
        return {
            'active_page': 'sos',
            **super().get_context_data(*args, **kwargs),
            'events': Event.objects.manageable_by_user(user=self.request.user, conference=self.conference, staff_can_manage=False)
            .filter(kind=Event.Kind.SELF_ORGANIZED)
            .select_related('owner'),
            'projects': Project.objects.manageable_by_user(self.conference, user=self.request.user, staff_can_manage=False).filter(assembly=None),
        }


class SoSCreateView(SoSFormMixin, ConferenceLoginRequiredMixin, CreateView):
    def dispatch(self, request, *args, **kwargs):
        if self.conference.self_organized_sessions_assembly_id is None:
            messages.error(self.request, _('SoS__missing_assembly'))
            logging.error('No SOS assembly set on conference "%(conference)s"!', {'conference': self.conference.slug})
            return redirect('backoffice:sos')

        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self, *args, **kwargs):
        return {'create': True, **super().get_form_kwargs(*args, **kwargs)}


class SoSUpdateView(EventPublicationMixin, SoSFormMixin, ConferenceLoginRequiredMixin, UpdateView):
    def get_queryset(self, *args, **kwargs):
        return (
            Event.objects.manageable_by_user(user=self.request.user, conference=self.conference).filter(kind=Event.Kind.SELF_ORGANIZED).select_related('owner')
        )


class SoSDeleteView(ConferenceLoginRequiredMixin, DeleteView):
    def get_queryset(self, *args, **kwargs):
        return (
            Event.objects.manageable_by_user(user=self.request.user, conference=self.conference).filter(kind=Event.Kind.SELF_ORGANIZED).select_related('owner')
        )

    def form_valid(self, form):
        response = super().form_valid(form)
        messages.success(self.request, _('SoS__removed %(event_name)s') % {'event_name': self.object.name})
        logger.info('Event (sos) "%(name)s" (%(pk)s) removed by <%(user)s>', {'name': self.object.name, 'pk': self.object.pk, 'user': self.request.user})
        ConferenceExportCache.signal_schedule_modification(self.conference)
        return response

    def get_success_url(self, *args, **kwargs):
        return reverse('backoffice:sos')
