import logging
import re
from urllib.parse import urlparse

from django.conf import settings
from django.db.models import Q
from django.urls import Resolver404, resolve, reverse
from django.utils.translation import gettext_lazy as _

from core.models import Conference, PlatformUser

from backoffice.views.mixins import ConferenceLoginRequiredMixin, guess_active_sidebar_item

logger = logging.getLogger(__name__)


RE_UUID = re.compile(r'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}')


def lookup_moderation_items(conference: Conference, query: str):
    # search uuids in URL
    uuids_in_query = RE_UUID.findall(query)
    words = [x for x in query.split() if x not in uuids_in_query]

    candidates = []

    # if the query is an URL try to map the object directly
    try:
        if query.startswith(settings.PLAINUI_BASE_URL):
            query_url = query[len(settings.PLAINUI_BASE_URL) :]
        elif query.startswith(settings.BACKOFFICE_BASE_URL):
            query_url = query[len(settings.BACKOFFICE_BASE_URL) :]
        else:
            query_url = urlparse(query).path

        p = resolve(query_url)
        if '/assembly' in p.route:
            if 'pk' in p.kwargs:
                candidates.append(('assembly', p.kwargs['pk']))
            for kw in ['slug', 'assembly_slug']:
                if kw in p.kwargs:
                    for a in conference.assemblies.filter(slug=p.kwargs[kw]).values('pk'):
                        candidates.append(('assembly', a['pk']))
        if '/event' in p.route:
            if 'pk' in p.kwargs:
                candidates.append(('event', p.kwargs['pk']))
            if 'slug' in p.kwargs:
                for e in conference.events.filter(slug=p.kwargs['slug']).values('pk'):
                    candidates.append(('event', e['pk']))
        if '/badge' in p.route:
            if 'pk' in p.kwargs:
                candidates.append(('badge', p.kwargs['pk']))
            if 'token' in p.kwargs:
                for b in conference.badges.filter(tokens__token=p.kwargs['token']).values('pk'):
                    candidates.append(('badge', b['pk']))
        if '/board' in p.route:
            for kw in ['id', 'pk']:
                if kw in p.kwargs:
                    candidates.append(('board', p.kwargs[kw]))
        if '/project' in p.route:
            if 'pk' in p.kwargs:
                candidates.append(('project', p.kwargs['pk']))
            if 'token' in p.kwargs:
                for b in conference.projects.filter(tokens__token=p.kwargs['slug']).values('pk'):
                    candidates.append((' ', b['pk']))

    except Resolver404:
        pass

    # if there are UUIDs found cheat a little: check for keywords in URL instead of proper matching
    if uuids_in_query:
        if 'assembly' in query:
            for a in conference.assemblies.filter(pk__in=uuids_in_query).values('pk'):
                candidates.append(('assembly', a['pk']))
        if 'event' in query:
            for e in conference.events.filter(pk__in=uuids_in_query).values('pk'):
                candidates.append(('event', e['pk']))
        if 'badge' in query:
            for b in conference.badges.filter(Q(pk__in=uuids_in_query) | Q(tokens__id__in=uuids_in_query)).values('pk'):
                candidates.append(('badge', b['pk']))
        if 'wiki' in query or 'page' in query:
            for p in conference.pages.filter(pk__in=uuids_in_query).values('pk'):
                candidates.append(('wiki', p['pk']))
        if 'user' in query:
            for u in conference.users.filter(user_id__in=uuids_in_query).values('pk'):
                candidates.append(('user', u['pk']))
        if 'project' in query:
            for u in conference.projects.filter(user_id__in=uuids_in_query).values('pk'):
                candidates.append(('project', u['pk']))

    # trigger on keywords if no UUID was found
    else:
        q_slug = Q()
        q_name = Q()
        q_username = Q()
        q_display_name = Q()

        for word in words:
            q_slug = q_slug & Q(slug__icontains=word)
            q_name = q_name & Q(name__icontains=word)
            q_username = q_username & Q(username__icontains=word)
            q_display_name = q_display_name & Q(display_name__icontains=word)

        for a in conference.assemblies.filter(q_slug).values('pk'):
            candidates.append(('assembly', a['pk']))
        for b in conference.badges.filter(q_name).values('pk'):
            candidates.append(('badge', b['pk']))
        for e in conference.events.filter(q_slug).values('pk'):
            candidates.append(('event', e['pk']))
        for u in PlatformUser.objects.filter(conferences__conference=conference).filter(q_username | q_display_name | q_slug).values('pk'):
            candidates.append(('user', u['pk']))
        for p in conference.pages.filter(q_slug).values('pk'):
            candidates.append(('wiki', p['pk']))
        for proj in conference.projects.filter(q_slug).values('pk'):
            candidates.append(('project', proj['pk']))

    return {
        'assemblies': conference.assemblies.filter(pk__in=[pk for t, pk in candidates if t == 'assembly']),
        'events': conference.events.filter(pk__in=[pk for t, pk in candidates if t == 'event']),
        'badges': conference.badges.filter(pk__in=[pk for t, pk in candidates if t == 'badge']),
        'board': conference.bulletin_board_entries.filter(pk__in=[pk for t, pk in candidates if t == 'board']),
        'project': conference.projects.filter(pk__in=[pk for t, pk in candidates if t == 'project']),
        'wiki': conference.pages.filter(pk__in=[pk for t, pk in candidates if t == 'wiki']),
        'users': PlatformUser.objects.filter(conferences__conference=conference).filter(pk__in=[pk for t, pk in candidates if t == 'user']),
    }


class ModerationAdminMixin(ConferenceLoginRequiredMixin):
    permission_required = ['core.moderation']

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['active_page'] = 'moderation'

        context['sidebar'] = {
            'title': 'Moderation',
            'title_link': reverse('backoffice:moderation'),
            'items': [
                {
                    'caption': _('Assemblies'),
                    'link': reverse('backoffice:moderation-assembly-list'),
                    'count': self.conference.assemblies.count(),
                    'expanded': False,
                },
                {
                    'caption': _('Events'),
                    'link': reverse('backoffice:moderation-event-list'),
                    'expanded': False,
                    'count': self.conference.events.count(),
                },
                {
                    'caption': _('BulletinBoardEntries'),
                    'link': reverse('backoffice:moderation-board-list'),
                    'expanded': False,
                    'count': self.conference.bulletin_board_entries.count(),
                },
                {
                    'caption': _('PlatformUsers'),
                    'link': reverse('backoffice:moderation-user-list'),
                    'expanded': False,
                    'count': self.conference.users.count(),
                },
                {
                    'caption': _('Projects'),
                    'link': reverse('backoffice:moderation-project-list'),
                    'expanded': False,
                    'count': self.conference.projects.count(),
                },
                {
                    'caption': _('Badges'),
                    'link': reverse('backoffice:moderation-badge-list'),
                    'expanded': False,
                    'count': self.conference.badges.count(),
                },
                {
                    'caption': _('StaticPages'),
                    'link': reverse('backoffice:moderation-wiki-list'),
                    'expanded': False,
                    'count': self.conference.pages.count(),
                },
            ],
        }

        # try to guess 'active' sidebar item
        guess_active_sidebar_item(self.request, context['sidebar']['items'], with_query_string=True)

        return context
