from django.contrib import messages
from django.http import JsonResponse
from django.shortcuts import redirect
from django.urls import reverse
from django.views import View
from django.views.generic import CreateView, DeleteView, DetailView, ListView, TemplateView, UpdateView
from django.views.generic.detail import SingleObjectMixin

from core.models import ScheduleSource, ScheduleSourceImport
from core.schedules import ScheduleTypeManager

from backoffice.forms import ScheduleSourceCreateForm
from backoffice.views.mixins import ConferenceLoginRequiredMixin


class ScheduleAdminMixin(ConferenceLoginRequiredMixin):
    permission_required = ['core.scheduleadmin']

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['active_page'] = 'schedules'
        return ctx


class SchedulesIndexView(ScheduleAdminMixin, TemplateView):
    template_name = 'backoffice/schedules_index.html'

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['sources'] = self.conference.schedule_sources.select_related('assembly').all().order_by('assembly__name')
        return ctx


class ScheduleSourcesListView(ScheduleAdminMixin, ListView):
    template_name = 'backoffice/schedule_source-list.html'
    model = ScheduleSource

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['import_types'] = {t: str(k) for t, k in ScheduleTypeManager.types.items()}
        return ctx


class ScheduleSourcesDetailView(ScheduleAdminMixin, DetailView):
    template_name = 'backoffice/schedule_source-detail.html'
    model = ScheduleSource

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['imports'] = ctx['object'].imports.all()[:11]
        ctx['has_more_imports'] = len(ctx['imports']) > 10
        ctx['imports'] = ctx['imports'][:10]
        return ctx


class ScheduleSourcesDoImportView(ScheduleAdminMixin, SingleObjectMixin, View):
    model = ScheduleSource

    def post(self, request, *args, **kwargs):
        src = self.get_object()
        job = ScheduleSourceImport(schedule_source=src, state=ScheduleSourceImport.State.PREPARED)
        job.save()
        if src.assembly is not None:
            messages.info(request, f"+ ScheduleSourceImport {job.pk} for '{src.assembly.slug}' ({src.pk})")
        else:
            messages.info(request, f'+ ScheduleSourceImport {job.pk} for wildcard assembly ({src.pk})')

        try:
            result = job.do_import()
            if result:
                messages.success(request, f'ScheduleSourceImport {job.pk} succeeded')
            else:
                messages.warning(request, f'ScheduleSourceImport {job.pk} failed')
        except Exception as err:
            messages.error(request, f'ScheduleSourceImport {job.pk} threw exception: {err}')

        return redirect('backoffice:schedulesourceimport-detail', pk=str(job.pk))


class ScheduleSourcesCreateView(ScheduleAdminMixin, CreateView):
    model = ScheduleSource
    form_class = ScheduleSourceCreateForm

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['conference'] = self.conference
        return kwargs

    def form_valid(self, form):
        form.instance.save()

        if form.instance.assembly:
            messages.success(self.request, f'+ ScheduleSource on "{form.instance.assembly.slug}": {form.instance.id}')
        else:
            messages.success(self.request, f'+ ScheduleSource on wildcard assembly: {form.instance.id}')
        return redirect('backoffice:schedulesource-detail', pk=str(form.instance.id))

    def form_invalid(self, form):
        messages.error(self.request, 'Error in form, please try again: ' + str(form.errors))
        return redirect('backoffice:schedulesource-list')


class ScheduleSourcesUpdateView(ScheduleAdminMixin, UpdateView):
    model = ScheduleSource
    fields = ['import_type', 'import_frequency', 'import_timeout', 'import_url', 'import_configuration']
    template_name = 'backoffice/schedule_source-form.html'

    def get_success_url(self):
        return reverse('backoffice:schedulesource-detail', kwargs={'pk': str(self.object.id)})


class ScheduleSourcesDeleteView(ScheduleAdminMixin, DeleteView):
    model = ScheduleSource
    template_name = 'backoffice/schedule_source-confirm_delete.html'

    def get_success_url(self):
        return reverse('backoffice:schedulesource-list')


class ScheduleSourceImportDetailView(ScheduleAdminMixin, DetailView):
    template_name = 'backoffice/schedule_source_import-detail.html'
    model = ScheduleSourceImport

    def get_context_data(self, *args, **kwargs):
        action2sort = {'error': 10, 'removed': 20, 'missing': 25, 'added': 30, 'seen': 80}
        type2sort = {'room': 10, 'event': 20, 'speaker': 50}
        ctx = super().get_context_data(*args, **kwargs)
        if data := ctx['object'].data:
            ctx['activity'] = sorted(
                data.get('_activity', []),
                key=lambda item: (action2sort.get(item['action'], 99), type2sort.get(item['type'], 99), item['source_id']),
            )
        else:
            ctx['activity'] = None
        return ctx


class ScheduleSourceImportDataView(ScheduleAdminMixin, DetailView):
    model = ScheduleSourceImport

    def get(self, *args, **kwargs):
        return JsonResponse(self.get_object().data, json_dumps_params={'indent': 2})
