from typing import Any

from django.contrib.postgres.aggregates.general import ArrayAgg
from django.db.models import BooleanField, Case, Count, When
from django.db.models.query import Q, QuerySet
from django.http import HttpRequest, HttpResponse
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext as _
from django.views.generic import DetailView, FormView, ListView
from django.views.generic.edit import CreateView, DeleteView, UpdateView
from rules.contrib.views import AutoPermissionRequiredMixin

from core.forms import TeamForm
from core.models import ConferenceMember, Invitation, Team
from core.views.mixins import FormMesssageMixin

from backoffice.views.mixins import ConferenceRuleLoginRequiredMixin, SingleUUIDObjectMixin, guess_active_sidebar_item


class TeamNavContextMixin(ConferenceRuleLoginRequiredMixin):
    def get_context_data(self, **kwargs) -> dict[str, Any]:
        member = ConferenceMember.get_member(
            conference=self.conference,
            user=self.request.user,
        )
        teams = [
            {
                'caption': team['name'],
                'link': reverse(
                    'backoffice:team',
                    kwargs={
                        'uuid': team['uuid'],
                    },
                ),
                'classes': [],
            }
            for team in Team.objects.all().values(
                'uuid',
                'name',
                'require_staff',
            )
            if member.is_staff or not team['require_staff']
        ]
        team_entry = {
            'caption': _('Teams'),
            'link': reverse('backoffice:teams'),
            'children': teams,
            'count': len(teams),
            'expanded': True,
        }

        context = {
            **super().get_context_data(**kwargs),
            'active_page': 'teams',
            'sidebar': {
                'title': _('Teams'),
                'title_link': reverse('backoffice:teams'),
                'items': [
                    team_entry,
                ],
            },
        }

        guess_active_sidebar_item(self.request, context['sidebar']['items'])

        return context


class TeamListView(TeamNavContextMixin, ListView):
    model = Team
    template_name = 'backoffice/teams/list.html'
    context_object_name = 'teams'
    permission_required = 'core.view_team'

    def get_queryset(self) -> QuerySet[Team]:
        member = ConferenceMember.get_member(
            conference=self.conference,
            user=self.request.user,
        )
        qs = super().get_queryset().order_by('name')
        if member.is_staff:
            qs = qs.annotate(
                members_count=Count('members'),
            )
        else:
            qs = qs.filter(require_staff=False)
        return qs.annotate(
            member_list=ArrayAgg('members__user'),
            self_manage_count=Count('members', filter=Q(members__user=self.request.user, members__can_manage=True)),
        ).annotate(
            can_manage=Case(When(self_manage_count__gt=0, then=True), default=False, output_field=BooleanField()),
        )


class TeamDetailView(SingleUUIDObjectMixin, TeamNavContextMixin, AutoPermissionRequiredMixin, DetailView):
    model = Team
    template_name = 'backoffice/teams/detail.html'
    context_object_name = 'team'

    def get_queryset(self) -> QuerySet[Team]:
        return super().get_queryset().prefetch_related('members__user')

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        if not Team.type_is(team := self.object):  # pragma: no cover
            raise ValueError('Invalid object type')
        member_id = self.get_queryset().filter(uuid=team.uuid, members__user=self.request.user).values_list('members__id', flat=True).first()
        log_entries = team.logentries.order_by('-timestamp')
        return {
            **super().get_context_data(**kwargs),
            'member_id': member_id,
            'received_invitations': team.received_invitations.filter(state=Invitation.RequestsState.REQUESTED),
            'sent_invitations': team.sent_invitations.filter(state=Invitation.RequestsState.REQUESTED),
            'all_log_entries': log_entries.all(),
            'mst_log_entries': log_entries.filter(comment__isnull=False),
        }


class TeamFormMixin(AutoPermissionRequiredMixin, TeamNavContextMixin, FormMesssageMixin, FormView):
    model = Team
    object_name = 'team'
    form_class = TeamForm
    template_name = 'backoffice/teams/create_edit.html'

    def get_success_url(self):
        return reverse('backoffice:team', kwargs={'uuid': self.object.uuid})

    def get_form_kwargs(self) -> dict[str, Any]:
        return {
            **super().get_form_kwargs(),
            'super_user': self.request.user.is_superuser,
            'conference': self.conference,
        }


class TeamCreateView(TeamFormMixin, CreateView):
    success_message = _('Team__create__success %(name)s')


class TeamUpdateView(SingleUUIDObjectMixin, TeamFormMixin, UpdateView):
    success_message = _('Team__update__success %(name)s')


class TeamDeleteView(SingleUUIDObjectMixin, FormMesssageMixin, TeamNavContextMixin, AutoPermissionRequiredMixin, DeleteView):
    model = Team
    template_name = 'backoffice/components/confirmation_modal.html'
    success_url = reverse_lazy('backoffice:teams')
    success_message = _('Team__delete__success')

    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        if request.POST.get('confirmation', request.GET.get('confirmation', 'false')) != 'true':
            return self.get(request, *args, **kwargs)
        return super().post(request, *args, **kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        return {
            **super().get_context_data(**kwargs),
            'confirmation_title': _('Team__delete__warning__header'),
            'confirmation_body': _('Team__delete__warning__text %(team)s') % {'team': self.object.name},
            'confirmation_class': 'danger',
            'confirmation_submit': _('Team__delete__submit'),
            'confirmation_cancel_url': reverse('backoffice:team', kwargs={'uuid': self.object.uuid}),
        }
