import logging
from typing import TYPE_CHECKING, Generic, TypeVar

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models import Model, QuerySet

from core.models.conference import ConferenceMember

if TYPE_CHECKING:
    from core.models.conference import Conference
    from core.models.users import PlatformUser

_ModelType = TypeVar('_ModelType', bound=Model)

logger = logging.getLogger(__name__)


class ConferenceManagerMixin(models.Manager, Generic[_ModelType]):
    """A mixin for managers that adds utility functions for conference and assembly queryset filtering.

    Args:
        Generic (_type_): The type of the model that the manager is used for.

    properties:
        staff_permissions (list[str] | None): A list of staff permissions that are required by staff to manage the queryset.
                                              If None, all staff members are allowed.
        conference_filter (str): The name of the field that is used to filter the queryset by conference.
        assembly_filter (str | None): The name of the field that is used to filter the queryset by assembly. If None, no assembly filtering is performed.
    """

    staff_permissions: list[str] | None = None
    conference_filter: str = 'conference'
    assembly_filter: str | None = None
    select_related_fields: list[str] | None = None
    prefetch_related_fields: list[str] | None = None

    def apply_public_filter(self, queryset: QuerySet[_ModelType], member: ConferenceMember | None = None) -> QuerySet[_ModelType]:
        """Returns a queryset of objects that are accessible for public consumption.

        Args:
            queryset (QuerySet[_ModelType]): The current queryset.
            member (ConferenceMember): The member that is trying to access the queryset.

        Returns:
            QuerySet[_ModelType]: A queryset of objects that are accessible for public consumption.
        """
        raise NotImplementedError('apply_public_filter must be implemented in subclasses')

    def apply_assembly_rights_filter(
        self,
        queryset: QuerySet[_ModelType],
        member: ConferenceMember,
        *,
        only_manageable: bool | None = None,
    ) -> QuerySet[_ModelType]:
        """Returns a queryset of objects that are manageable for assemblies where the given member is a member.

        Args:
            queryset (QuerySet[_ModelType]): The current queryset.
            member (ConferenceMember): The member that is trying to access the queryset.
        Kwargs:
            only_manageable (bool): If True, only objects that the user can manage are returned. Defaults to False.

        Returns:
            QuerySet[_ModelType]: A Queryset of objects that the user is allowed to manage.
        """
        if only_manageable is None:
            raise ValueError('management_permission_required must be set')
        if self.assembly_filter is None:
            raise NotImplementedError('Either assembly_filter must be set for the subclasses or apply_assembly_rights_filter must be implemented')
        if self.assembly_filter == 'self':
            prefix = ''
        else:
            prefix = f'{self.assembly_filter}__'
            queryset = queryset.select_related(self.assembly_filter)

        from core.models.assemblies import Assembly, AssemblyMember  # pylint: disable=C0415 # Circular import avoidance

        allowed_members = AssemblyMember.objects.filter(member=member.user)
        allowed_members = allowed_members.filter(can_manage_assembly=True) if only_manageable else allowed_members
        queryset = queryset.filter(**{f'{prefix}members__in': allowed_members}).exclude(**{f'{prefix}state__in': [Assembly.State.HIDDEN]})
        return queryset

    def apply_self_organized_rights_filter(
        self,
        queryset: QuerySet[_ModelType],
        member: ConferenceMember,
        *,
        only_manageable: bool | None = None,
    ) -> QuerySet[_ModelType]:
        """Returns a queryset of objects that are manageable in a self_organized context where the given member is allowed to view/manage.
        If this is not implemented in the subclass, the queryset will be empty.

        Args:
            queryset (QuerySet[_ModelType]): The current queryset.
            member (ConferenceMember): The member that is trying to access the queryset.
        Kwargs:
            management_permission_required (bool): If True, only objects that the user can manage are returned. Defaults to False.

        Returns:
            QuerySet[_ModelType]: A Queryset of objects that the user is allowed to manage.
        """
        return queryset.none()

    def get_conference_queryset(self, conference: 'Conference') -> QuerySet[_ModelType]:
        """
        Returns a queryset of objects that belong to a given conference.

        Args:
            conference (Conference): The conference which will be used to look up objects.

        Returns:
            QuerySet[_ModelType]: A Queryset of objects that belong to the given conference.
        """
        if conference is None:
            raise ValueError('conference must be set')
        if self.conference_filter is None:
            raise ImproperlyConfigured('conference_filter must be set')
        return self.get_queryset().filter(**{self.conference_filter: conference})

    def conference_accessible(self, conference: 'Conference', member: ConferenceMember | None = None) -> QuerySet[_ModelType]:
        """
        Returns a queryset of objects that are accessible for a given conference.

        Args:
            conference (Conference): The conference which will be used to look up objects.
            member (ConferenceMember | None, optional): The member that is trying to access the queryset. Defaults to None.

        Returns:
            QuerySet[_ModelType]: A Queryset of objects that are accessible for the given conference.
        """
        if not conference.is_published:
            return self.get_queryset().none()
        if isinstance(member, ConferenceMember):
            return self.apply_public_filter(self.get_conference_queryset(conference), member)
        return self.apply_public_filter(self.get_conference_queryset(conference))

    def _get_associated_with_user(
        self,
        conference: 'Conference',
        *,
        user: 'PlatformUser | None' = None,
        member: ConferenceMember | None = None,
        only_manageable: bool | None = None,
        staff_can_manage: bool = True,
        show_public: bool = False,
    ) -> QuerySet[_ModelType]:
        if not user and not member:
            raise ValueError('Either user or member must be given')
        if member and user and user != member.user:
            raise ValueError('Member does not match given user')
        if only_manageable is None:
            raise ValueError('management_permission_required must be set')
        member = ConferenceMember.get_member(conference, user=user, member=member)
        qs = self.get_conference_queryset(conference)
        if not isinstance(member, ConferenceMember):
            return self.conference_accessible(conference=conference, member=member) if show_public else qs.none()
        if member.is_staff and staff_can_manage:
            if self.staff_permissions is None or (self.staff_permissions and member.has_perms(*self.staff_permissions, require_staff=True)):
                return qs
        id_list = (
            self.apply_assembly_rights_filter(qs, member, only_manageable=only_manageable)
            .values('id')
            .union(self.apply_self_organized_rights_filter(qs, member, only_manageable=only_manageable).values('id'))
        )
        if show_public:
            id_list = id_list.union(self.apply_public_filter(qs, member).values('id'))
        return qs.filter(id__in=id_list).select_related(*self.select_related_fields or []).prefetch_related(*self.prefetch_related_fields or [])

    def associated_with_user(
        self,
        conference: 'Conference',
        *,
        user: 'PlatformUser | None' = None,
        member: ConferenceMember | None = None,
        staff_can_see: bool = True,
        show_public: bool = False,
    ) -> QuerySet[_ModelType]:
        """Returns a queryset of objects that are accessible by the given user.

        Args:
            conference (Conference): The conference which will be used to look up objects.
        Kwargs:
            user (PlatformUser | None, optional): Look up objects for this user. Optional if member is given. Defaults to None.
            member (ConferenceMember | None, optional): Look up objects for this uer. Optional if user is given. Defaults to None.
            management_permission_required (bool): If True, only objects that the user is associated are returned. Defaults to None.
            staff_can_see (bool): If True, staff members with the appropriate permissions can see all objects. Defaults to True.

        Raises:
            ValueError: Raised if the given parameters are not valid.

        Returns:
            QuerySet[_ModelType]: A Queryset of objects that are accessible by the given user.
        """
        return self._get_associated_with_user(
            conference,
            user=user,
            member=member,
            only_manageable=False,
            staff_can_manage=staff_can_see,
            show_public=show_public,
        )

    def manageable_by_user(
        self,
        conference: 'Conference',
        *,
        user: 'PlatformUser | None' = None,
        member: ConferenceMember | None = None,
        staff_can_manage: bool = True,
    ) -> QuerySet[_ModelType]:
        """Returns a queryset of objects that are accessible by the given user.

        Args:
            conference (Conference): The conference which will be used to look up objects.
        Kwargs:
            user (PlatformUser | None, optional): Look up objects for this user. Optional if member is given. Defaults to None.
            member (ConferenceMember | None, optional): Look up objects for this user. Optional if user is given. Defaults to None.
            staff_can_manage (bool): If True, staff members with the appropriate permissions can manage all objects. Defaults to True.

        Raises:
            ValueError: Raised if the given parameters are not valid.

        Returns:
            QuerySet[_ModelType]: A Queryset of objects that are manageable by the given user.
        """
        return self._get_associated_with_user(
            conference,
            user=user,
            member=member,
            only_manageable=True,
            staff_can_manage=staff_can_manage,
        )
