import logging
import re
import uuid

from urllib3.util import Url, parse_url

from django.conf import settings
from django.contrib.postgres import indexes as pg_indexes
from django.contrib.postgres import search as pg_search
from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.html import strip_tags
from django.utils.translation import gettext_lazy as _

from core.fields import ConferenceReference
from core.markdown import render_markdown_ex, store_relationships
from core.models.conference import Conference
from core.models.users import PlatformUser
from core.utils import GitRepo


class StaticPageNamespace(models.Model):
    class Meta:
        unique_together = [
            ('conference', 'prefix'),
        ]
        verbose_name = _('StaticPageNamespace')
        verbose_name_plural = _('StaticPageNamespaces')

    conference = ConferenceReference(related_name='wiki_namespaces')
    prefix = models.SlugField(
        help_text=_('StaticPageNamespace__prefix__help'),
        verbose_name=_('StaticPageNamespace__prefix'),
    )

    groups = ArrayField(
        models.CharField(max_length=100),
        help_text=_('StaticPageNamespace__groups__help'),
        verbose_name=_('StaticPageNamespace__groups'),
    )

    upstream_url = models.URLField(
        blank=True,
        null=True,
        help_text=_('StaticPageNamespace__upstream_url__help'),
        verbose_name=_('StaticPageNamespace__upstream_url'),
    )

    upstream_image_base_url = models.URLField(
        blank=True,
        null=True,
        help_text=_('StaticPageNamespace__upstream_image_base_url__help'),
        verbose_name=_('StaticPageNamespace__upstream_image_base_url'),
    )

    def clean(self):
        self.prefix = self.prefix.upper()

    def __str__(self):
        return self.prefix

    @staticmethod
    def _extract_doc_metadata(markup) -> tuple[dict[str, str], str]:
        if markup.startswith('---\n'):
            metadata_raw, content = markup[4:].split('---\n', maxsplit=1)
            metadata = {}
            for line in metadata_raw.splitlines():
                k, v = line.split(':', maxsplit=1)
                metadata[k.strip()] = v.strip()
        else:
            metadata = {}
            content = markup

        # TODO: optionally convert mkdocs-specific non-markdown syntax?

        return metadata, content

    def fetch_upstream(self):
        if not (url := self.upstream_url):
            return {}

        RE_MARKDOWN_IMAGE = re.compile(r'!\[(.+?)\]\((.+?)\)')

        with GitRepo(url) as repo:
            docs = repo.get_documents(glob='*.md')

            if image_base_url := self.upstream_image_base_url:
                image_base_url_url = parse_url(image_base_url)

                # replace relative image urls in documents
                def derive_image_url(match: re.Match):
                    image_url = match.group(2)
                    if '://' in image_url:
                        # seems to be a valid URL => don't change anything
                        return match.string

                    if image_url.startswith('./'):
                        image_url = image_url[2:]

                    if not image_url.startswith('/'):
                        new_image_url = image_base_url + image_url
                    else:
                        new_image_url = Url(
                            scheme=image_base_url_url.scheme,
                            auth=image_base_url_url.auth,
                            host=image_base_url_url.host,
                            port=image_base_url_url.port,
                            path=image_url,
                        )

                    return f'![{match.group(1)}]({new_image_url})'

                for doc_name, doc in docs.items():
                    docs[doc_name] = RE_MARKDOWN_IMAGE.sub(derive_image_url, doc)

        existing_pages = {page.slug: page for page in StaticPage.objects.filter(conference=self.conference, slug__istartswith=self.prefix)}
        obsolete_pages = set(existing_pages.keys())

        for doc_filename, document_raw in docs.items():
            page_slug = self.prefix.lower() + doc_filename.rstrip('.md')
            page = existing_pages.get(page_slug)  # type: StaticPage

            try:
                doc_metadata, document = self._extract_doc_metadata(document_raw)

                if page is not None:
                    # will update the existing page, note that we've seen it
                    obsolete_pages.remove(page_slug)
                    current_rev = page.revisions.get(revision=page.public_revision)
                    if current_rev.body != document:
                        # the document content has been changed
                        page_rev = page.revisions.create(
                            title=doc_metadata.get('title') or current_rev.title,
                            body=document,
                            is_draft=False,
                            author=None,
                            timestamp=timezone.now(),
                        )
                        page_rev.set_public()
                        logging.info('wiki page "%s" updated (old revision #%s, new #%s)', page_slug, current_rev.revision, page_rev.revision)
                        current_rev.delete()
                    else:
                        logging.debug('wiki page "%s" unchanged (revision #%s)', page_slug, current_rev.revision)

                else:
                    # create new page
                    page = StaticPage.objects.create(conference=self.conference, slug=page_slug)
                    page_rev = page.revisions.create(
                        title=doc_metadata.get('title') or page_slug[len(self.prefix) :].capitalize(),
                        body=document,
                        is_draft=False,
                        author=None,
                        timestamp=timezone.now(),
                    )
                    page_rev.set_public()
                    logging.info('wiki page "%s" created (revision #%s)', page_slug, page_rev.revision)
            except Exception:
                logging.exception('wiki page "%s" failed to import', page_slug)
                if page is not None:
                    # don't remove the previous version if importing a new one fails
                    obsolete_pages.remove(page_slug)

        updated = StaticPage.objects.filter(conference=self.conference).bulk_update(
            objs=(existing_pages[page_slug] for page_slug in obsolete_pages),
            fields={'privacy': StaticPage.Privacy.PERM},
        )
        if updated > 0:
            logging.info('set %s wiki page(s) "%s" non-public (removed upstream): %s', updated, page_slug, ';'.join(obsolete_pages))


class StaticPageManager(models.Manager):
    def accessible_by_user(self, user: PlatformUser, conference: Conference, language: str | None = ''):
        if language is not None:
            qs = self.get_queryset().filter(language=language, conference=conference)
        else:
            qs = self.get_queryset().filter(conference=conference)

        # for global staff apply no more limits
        if user.is_superuser or user.is_staff:
            return qs

        # content team can access non-public pages
        if user.has_conference_staff_permission(conference, 'core.static_pages'):
            return qs

        # return only pages which are marked public
        pages_filters = Q(privacy=StaticPage.Privacy.NONE)
        if user.is_in_conference(conference):
            pages_filters |= Q(privacy=StaticPage.Privacy.CONFERENCE)

        return qs.filter(public_revision__gt=0).filter(pages_filters)

    def get_editable_page(self, user: PlatformUser, conference: Conference, language: str, slug: str, check=False) -> tuple['StaticPage | None', bool]:
        """
        Request to edit the Static Page with slug `slug` in the context of `user` and `conference` and in the language `langguage`.

        Returns a tuple `(static_page, exists)` with `static_page` the `StaticPage` instance (or `None`) and `exists` a boolean iff the `StaticPage`
        already exists in the database and the user has sufficient permission to create it.

        `static_page` will be `None` iff the user is not allowed to edit this Page (even if it does not exist).
        Otherwise, it will be a `StaticPage` instance for the Page.
        If the Page does not yet exist in the database, the returned `StaticPage` is not stored in the database (`save()` needs to be called by the callee).
        In this case, `exists` will be `False`.
        """

        # Namespace check. Grant access if user has static page group for at least one matching namespace
        namespaces_matched = []  # type: List[StaticPageNamespace]
        namespace_slug = slug.upper()
        for namespace in StaticPageNamespace.objects.filter(conference=conference).all():
            if namespace_slug.startswith(namespace.prefix):
                namespaces_matched.append(namespace)

        group_check_failed = False
        if namespaces_matched:
            group_check_failed = True
            user_groups = set(user.get_staticpage_groups(conference) or [])
            for namespace in namespaces_matched:
                if set(namespace.groups) & user_groups != set():
                    group_check_failed = False
                    break

        qs = self.get_queryset().filter(conference=conference, slug=slug).filter(Q(language=language) | Q(language=None))
        if not check:
            qs = qs.select_for_update()
        sp = qs.first()

        # If Page does not exist, User can create it
        if sp is None:
            if group_check_failed and not user.has_conference_staff_permission(conference, 'core.static_pages'):
                return None, True

            new_page = StaticPage(
                conference=conference,
                slug=slug,
                language=language if settings.STATIC_PAGE_LOCALIZED_BY_DEFAULT else None,
            )
            return new_page, False

        # If Page is protected, check if user has staff permission to ignore it
        if group_check_failed or sp.protection == StaticPage.Protection.PERM:
            if not user.has_conference_staff_permission(conference, 'core.static_pages'):
                return None, True

        return sp, True

    def conference_accessible(self, conference: Conference, language: str):
        # fetch conference's pages with a public revision and a language set to the given one or no language set
        return self.get_queryset().filter(conference=conference, public_revision__gt=0).filter(Q(language=language) | Q(language=None))


class StaticPage(models.Model):
    class Protection(models.TextChoices):
        NONE = 'none', _('StaticPage__protection-none')
        PERM = 'perm', _('StaticPage__protection-perm')

    class Privacy(models.TextChoices):
        NONE = 'none', _('StaticPage__privacy-none')
        CONFERENCE = 'conf', _('StaticPage__privacy-conf')
        PERM = 'perm', _('StaticPage__privacy-perm')

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)

    conference = ConferenceReference(related_name='pages')
    slug = models.SlugField()
    language = models.CharField(max_length=20, blank=False, null=True, help_text=_('StaticPage__language__help'), verbose_name=_('StaticPage__language'))
    protection = models.CharField(
        max_length=20,
        default=Protection.NONE,
        choices=Protection.choices,
        help_text=_('StaticPage__protection__help'),
        verbose_name=_('StaticPage__protection'),
    )
    privacy = models.CharField(
        max_length=20, default=Privacy.NONE, choices=Privacy.choices, help_text=_('StaticPage__privacy__help'), verbose_name=_('StaticPage__privacy')
    )

    remove_html = models.BooleanField(
        default=True,
        help_text=_('StaticPage__remove_html__help'),
        verbose_name=_('StaticPage__remove_html'),
    )
    sanitize_html = models.BooleanField(
        default=True,
        help_text=_('StaticPage__sanitize_html__help'),
        verbose_name=_('StaticPage__sanitize_html'),
    )

    public_revision = models.PositiveIntegerField(default=0, help_text=_('StaticPage__public_revision__help'), verbose_name=_('StaticPage__public_revision'))
    title = models.CharField(max_length=200, help_text=_('StaticPage__title__help'), verbose_name=_('StaticPage__title'))
    body = models.TextField(default='', help_text=_('StaticPage__body__help'), verbose_name=_('StaticPage__body'))
    body_html = models.TextField(blank=True, null=True, editable=False, help_text=_('StaticPage__body_html__help'), verbose_name=_('StaticPage__body_html'))
    """HTML of the current public revision"""
    search_content = models.TextField(blank=True, null=True, editable=False)
    """content used for searching (internal use only)"""
    search_vector = pg_search.SearchVectorField(blank=True, editable=False, null=True)
    """PostgreSQL tsquery vector used for searching (internal use only, updated via trigger)"""

    objects = StaticPageManager()

    class Meta:
        constraints = [
            models.constraints.UniqueConstraint(
                fields=[
                    'conference',
                    'language',
                    'slug',
                ],
                name='unique_conferencestaticpage_slug',
            ),
        ]
        indexes = [
            pg_indexes.GinIndex(fields=['search_vector'], name='staticpage_searchvector'),
            models.Index(fields=['conference', 'language', 'privacy'], name='staticpage_conflangprivacy'),
        ]
        verbose_name = _('StaticPage')
        verbose_name_plural = _('StaticPages')

    def __str__(self):
        return f'{self.title}[{self.language}]'

    @cached_property
    def newest_revision(self):
        try:
            return self.revisions.order_by('-revision').first()
        except StaticPageRevision.DoesNotExist:
            return None

    def clean(self, *args, **kwargs):
        if self.public_revision > 0:
            try:
                StaticPageRevision.objects.get(page=self, revision=self.public_revision)
            except StaticPageRevision.DoesNotExist:
                raise ValidationError({'public_revision': _('StaticPage__public_revision__not_found')})
        else:
            self.search_content = None
            self.search_vector = None

    @property
    def is_localized(self):
        return self.language is not None

    @property
    def is_public_readable(self):
        return self.privacy in [self.Privacy.NONE, self.Privacy.CONFERENCE]

    @property
    def is_public_writeable(self):
        return self.protection == self.Protection.NONE


class StaticPageRevision(models.Model):
    page = models.ForeignKey(StaticPage, related_name='revisions', on_delete=models.CASCADE)
    revision = models.AutoField(primary_key=True)

    title = models.CharField(max_length=200, help_text=_('StaticPageRevision__title__help'), verbose_name=_('StaticPageRevision__title'))
    body = models.TextField(help_text=_('StaticPageRevision__body__help'), verbose_name=_('StaticPageRevision__body'))

    # `is_draft` wird momentan nicht mehr wirklich verwendet.
    # Um das wieder sinnvoll zu nutzen sollte eine Möglichkeit geschaffen zu werden, drafts
    # zu speichern, die dann zwar schon vorgeschrieben sind jedoch nicht veröffentlicht
    # werden. Und dann eine Möglichkeit einen draft zu veröffentlichen.
    # das Veröffentlichen sollte jedoch eine neue Revision erzeugen,
    # so dass Revisionen, die als `is_draft` markiert sind, dauerhaft `is_draft` bleiben.
    is_draft = models.BooleanField(default=False, help_text=_('StaticPageRevision__is_draft__help'), verbose_name=_('StaticPageRevision__is_draft'))

    author = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        help_text=_('StaticPageRevision__author__help'),
        verbose_name=_('StaticPageRevision__author'),
    )
    timestamp = models.DateTimeField(auto_now_add=True, help_text=_('StaticPageRevision__timestamp__help'), verbose_name=_('StaticPageRevision__timestamp'))

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['page', 'revision'], name='page_revision'),
        ]

    def __str__(self):
        return f'{self.page.conference.slug}/{self.page.slug} @ {self.revision}'

    @cached_property
    def is_public_revision(self):
        return self.page.public_revision == self.revision

    def set_public(self):
        """Sets this revision as the public version of the page."""
        assert not self.is_draft, 'Revision is marked as draft.'

        self.page.public_revision = self.revision
        self.page.title = self.title
        self.page.body = self.body
        render_result = render_markdown_ex(
            self.page.conference, self.body, allow_embedded_html=not self.page.remove_html, sanitize_html=self.page.sanitize_html
        )
        self.page.body_html = render_result.document
        store_relationships(self.page.conference, self.page, render_result)
        self.page.search_content = self.title + '\n' + strip_tags(self.page.body_html)
        self.page.search_vector = None
        self.page.save()
