from typing import TYPE_CHECKING, Any, TypeIs
from uuid import uuid4

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _
from rules.contrib.models import RulesModel
from rules.predicates import is_superuser, predicate

if TYPE_CHECKING:  # pragma: no cover
    from core.models import PlatformUser


class LastManagerError(ValidationError):
    pass


@predicate
def has_invitation(user: 'PlatformUser', team_member: 'TeamMember | None' = None) -> bool:
    if team_member is None:  # pragma: no cover
        return False
    return user.is_authenticated and team_member.team.sent_invitations.filter(requested_id=user.id).exists()


@predicate
def is_associated_user(user: 'PlatformUser', team_member: 'TeamMember | None' = None) -> bool:
    if team_member is None:  # pragma: no cover
        return False
    return user.is_authenticated and team_member.user == user


@predicate
def is_team_manager(user: 'PlatformUser', team_member: 'TeamMember | None' = None) -> bool:
    if team_member is None:  # pragma: no cover
        return False
    return user.is_authenticated and team_member.team.members.filter(user=user, can_manage=True).exists()


class TeamMember(RulesModel):
    class Meta:
        rules_permissions = {
            'add': is_team_manager | is_superuser | has_invitation,
            'change': is_team_manager | is_superuser,
            'delete': is_team_manager | is_superuser | is_associated_user,
        }
        unique_together = (('user', 'team'),)
        verbose_name = _('TeamMember')
        verbose_name_plural = _('TeamMembers')

    id = models.UUIDField(primary_key=True, default=uuid4, editable=False)
    user = models.ForeignKey('PlatformUser', related_name='teams', on_delete=models.CASCADE)
    team = models.ForeignKey('Team', related_name='members', on_delete=models.CASCADE)
    can_manage = models.BooleanField(
        default=False,
        verbose_name=_('TeamMember__can_manage'),
        help_text=_('TeamMember__can_manage__help'),
    )
    created = models.DateTimeField(auto_now_add=True, verbose_name=_('TeamMember__created'))
    updated = models.DateTimeField(auto_now=True, verbose_name=_('TeamMember__updated'))

    @classmethod
    def type_is(cls, obj: object) -> TypeIs['TeamMember']:
        return isinstance(obj, cls)

    def clean(self) -> None:
        if not self.can_manage and self.team.members.filter(can_manage=True).count() == 1:
            raise LastManagerError(_('TeamMember__clean__cannot_remove_last_manager'))
        return super().clean()

    def delete(self, using: Any = None, keep_parents: bool = False) -> tuple[int, dict[str, int]]:
        if self.can_manage and self.team.members.filter(can_manage=True).count() == 1:
            raise LastManagerError(_('TeamMember__delete__cannot_delete_last_manager'))
        return super().delete(using, keep_parents)
