from django.contrib.auth.models import Group, Permission
from django.core import mail
from django.test import TestCase, override_settings

from core.models.assemblies import Assembly, AssemblyMember
from core.models.conference import Conference, ConferenceMember
from core.models.users import PlatformUser, UserCommunicationChannel


class AssembliesTestsMixin(TestCase):
    def setUp(self):
        self.conference = Conference(slug='foo', name='Foo Conference', mail_footer='This is serious business w/ legal stuff.', is_public=True)
        self.conference.save()

        self.user_mgmt = PlatformUser(username='manager')
        self.user_mgmt.save()
        self.user_mgmt.communication_channels.create(
            channel=UserCommunicationChannel.Channel.MAIL,
            address='manager@unittest.local',
            is_verified=True,
            use_for_notifications=False,
        )
        self.user_mgmt.communication_channels.create(
            channel=UserCommunicationChannel.Channel.MAIL,
            address='notifications@unittest.local',
            is_verified=True,
            use_for_notifications=True,
        )
        self.user_mgmt2 = PlatformUser(username='manager2')
        self.user_mgmt2.save()
        self.user_mgmt2.user_permissions.add(Permission.objects.get(codename='assembly_team'))
        self.user_mgmt2.communication_channels.create(
            channel=UserCommunicationChannel.Channel.MAIL,
            address='notifications2@unittest.local',
            is_verified=True,
            use_for_notifications=True,
        )
        self.user_participant = PlatformUser(username='participant')
        self.user_participant.save()
        self.user_participant.communication_channels.create(
            channel=UserCommunicationChannel.Channel.MAIL,
            address='participant@unittest.local',
            is_verified=True,
            use_for_notifications=False,
        )
        self.user_visitor = PlatformUser(username='visitor')
        self.user_visitor.save()
        self.user_visitor.communication_channels.create(
            channel=UserCommunicationChannel.Channel.MAIL,
            address='visitor@unittest.local',
            is_verified=True,
            use_for_notifications=False,
        )
        self.user_anon = PlatformUser.get_anonymous_user()
        ConferenceMember(conference=self.conference, user=self.user_mgmt).save()
        ConferenceMember(conference=self.conference, user=self.user_mgmt2).save()
        ConferenceMember(conference=self.conference, user=self.user_participant).save()
        ConferenceMember(conference=self.conference, user=self.user_visitor).save()

        self.assembly1 = Assembly(conference=self.conference, slug='fnord', name='Fnord Assembly')
        self.assembly1.save()
        self.assembly1.members.create(member=self.user_mgmt, can_manage_assembly=True)
        self.assembly1.members.create(member=self.user_participant, can_manage_assembly=False, show_public=True)
        self.assembly2 = Assembly(conference=self.conference, slug='foo', name='Foo Bar')
        self.assembly2.members.create(member=self.user_mgmt2, can_manage_assembly=True)
        self.assembly2.save()


class AssembliesTests(AssembliesTestsMixin):
    @override_settings(SUPPORT_HTML_MAILS=True)
    def test_mail_plainhtml(self):
        self.assembly2.send_mail_to_managers(
            subject='Hello World',
            message='ASDF *FNORD*',
        )
        self.assertEqual(len(mail.outbox), 1)

        the_mail = mail.outbox[0]  # type: mail.EmailMultiAlternatives
        self.assertEqual(the_mail.subject, f'[{self.conference.slug}] {{{self.assembly2.name}}} Hello World')
        self.assertListEqual(the_mail.to, ['notifications2@unittest.local'])
        self.assertEqual(the_mail.from_email, 'assembly_team+mail_from_not_configured@localhost')
        self.assertListEqual(the_mail.reply_to, ['assembly_team+mail_from_not_configured@localhost'])

        # check that plaintext + html parts are present and contain the link
        self.assertEqual(the_mail.content_subtype, 'plain')
        self.assertEqual(len(the_mail.alternatives), 1)
        self.assertEqual(the_mail.alternatives[0][1], 'text/html')
        self.assertIn(self.assembly2.slug, the_mail.body)
        self.assertIn('FNORD', the_mail.body)
        self.assertIn(self.assembly2.slug, the_mail.alternatives[0][0])
        self.assertIn('FNORD', the_mail.alternatives[0][0])

    @override_settings(SUPPORT_HTML_MAILS=True)
    @override_settings(DEREFERRER_GLOBAL_ALLOWLIST=['https://events.ccc.de'])
    def test_mail_htmlwithlinks(self):
        self.assembly1.send_mail_to_managers(
            subject='Hello World',
            message='Please check [the blog](https://events.ccc.de/) and do stuff.\n' "And don't forget to visit [events.ccc.de](https://events.ccc.de/)!",
        )
        self.assertEqual(len(mail.outbox), 2)

        the_mail = mail.outbox[0]  # type: mail.EmailMultiAlternatives
        self.assertEqual(the_mail.subject, f'[{self.conference.slug}] {{{self.assembly1.name}}} Hello World')

        # check that plaintext + html parts are present and contain the link
        self.assertEqual(the_mail.content_subtype, 'plain')
        self.assertEqual(len(the_mail.alternatives), 1)
        self.assertEqual(the_mail.alternatives[0][1], 'text/html')
        self.assertIn(self.assembly1.slug, the_mail.body)
        self.assertIn('https://events.ccc.de/', the_mail.body)
        self.assertIn(self.assembly1.slug, the_mail.alternatives[0][0])
        self.assertIn('https://events.ccc.de/', the_mail.alternatives[0][0])


class AssembliesObjectManagerTests(AssembliesTestsMixin):
    fixtures = [
        'bootstrap_auth_groups',
    ]

    def setUp(self):
        super().setUp()

        self.conference2 = Conference(slug='bar', name='Bar Conference')
        self.conference2.save()
        self.assembly3 = Assembly(conference=self.conference2, slug='fnord', name='Fnord Assembly')
        self.assembly3.save()
        self.user_a3_mgmt = PlatformUser(username='manager3')
        self.user_a3_mgmt.save()
        ConferenceMember(conference=self.conference2, user=self.user_a3_mgmt).save()
        self.assembly3.members.create(member=self.user_a3_mgmt, can_manage_assembly=True, show_public=True)
        self.user_a3_part = PlatformUser(username='participant3')
        self.user_a3_part.save()
        ConferenceMember(conference=self.conference2, user=self.user_a3_part).save()
        self.assembly3.members.create(member=self.user_a3_part, can_manage_assembly=False, show_public=True)

        self.assembly_group = Group.objects.get(name='Assembly-Team')

        self.user_conf_staff = PlatformUser(username='conf_staff')
        self.user_conf_staff.save()
        self.cm_conf_staff = ConferenceMember(
            conference=self.conference,
            user=self.user_conf_staff,
            is_staff=True,
        )
        self.cm_conf_staff.save()
        self.cm_conf_staff.permission_groups.add(self.assembly_group)

        self.public_members = list(AssemblyMember.objects.filter(show_public=True))

    def test_associated_with_user(self):
        self.assertListEqual(
            list(AssemblyMember.objects.associated_with_user(user=self.user_conf_staff, conference=self.conference)),
            list(AssemblyMember.objects.filter(assembly__conference=self.conference)),
        )
        self.assertListEqual(
            list(AssemblyMember.objects.associated_with_user(user=self.user_conf_staff, conference=self.conference, staff_can_see=False)),
            [],
        )
        self.assertListEqual(
            list(AssemblyMember.objects.associated_with_user(user=self.user_visitor, conference=self.conference)),
            [],
        )

    def test_conference_accessible(self):
        self.assertListEqual(
            list(AssemblyMember.objects.conference_accessible(conference=self.conference)),
            list(AssemblyMember.objects.filter(assembly__conference=self.conference, show_public=True)),
        )
        self.assertListEqual(
            list(AssemblyMember.objects.conference_accessible(conference=self.conference)),
            list(AssemblyMember.objects.filter(assembly__conference=self.conference, show_public=True)),
        )
