from datetime import UTC, datetime, timedelta
from unittest.mock import patch

from django.test import TestCase
from django.utils import timezone

from core.models.conference import Conference, ConferenceDay


class ConferenceTests(TestCase):
    def setUp(self):
        self.conference = Conference(slug='foo', name='Foo Conference', is_public=True)
        self.conference.save()

    def test_get_conferenceday(self):
        self.conference.start = datetime(2042, 12, 27, 10, 00, tzinfo=UTC)
        self.conference.end = datetime(2042, 12, 30, 16, 00, tzinfo=UTC)

        self.assertEqual(self.conference.get_conference_day(datetime(2042, 12, 24, 18, 00, tzinfo=UTC)), -2)
        self.assertEqual(self.conference.get_conference_day(datetime(2042, 12, 26, 12, 34, tzinfo=UTC)), 0)
        self.assertEqual(self.conference.get_conference_day(datetime(2042, 12, 27, 3, 45, tzinfo=UTC)), 0)
        self.assertEqual(self.conference.get_conference_day(datetime(2042, 12, 27, 12, 34, tzinfo=UTC)), 1)
        self.assertEqual(self.conference.get_conference_day(datetime(2042, 12, 30, 2, 34, tzinfo=UTC)), 3)
        self.assertEqual(self.conference.get_conference_day(datetime(2042, 12, 30, 12, 34, tzinfo=UTC)), 4)
        self.assertEqual(self.conference.get_conference_day(datetime(2042, 12, 31, 5, 30, tzinfo=UTC)), 4)
        self.assertEqual(self.conference.get_conference_day(datetime(2042, 12, 31, 5, 30, tzinfo=UTC)), 4)

    def test_days_in(self):
        with patch.object(timezone, 'now', return_value=datetime(2042, 12, 27, 12, 34, 0, 0, tzinfo=UTC)):
            self.conference.start = timezone.now() - timedelta(days=1)
            self.conference.end = timezone.now() + timedelta(days=3)
            self.assertTrue(self.conference.is_running)
            self.assertFalse(self.conference.has_ended)
            self.assertEqual(5, self.conference.days_count)

            self.assertEqual(2, self.conference.current_day)

    def test_days_future(self):
        with patch.object(timezone, 'now', return_value=datetime(2042, 12, 27, 12, 34, 0, 0, tzinfo=UTC)):
            self.conference.start = timezone.now() + timedelta(days=10)
            self.conference.end = timezone.now() + timedelta(days=14)
            self.assertFalse(self.conference.is_running)
            self.assertFalse(self.conference.has_ended)
            self.assertEqual(5, self.conference.days_count)

            self.assertEqual(-9, self.conference.current_day)

    def test_days_past(self):
        with patch.object(timezone, 'now', return_value=datetime(2042, 12, 27, 12, 34, 0, 0, tzinfo=UTC)):
            self.conference.start = timezone.now() - timedelta(days=5)
            self.conference.end = timezone.now() - timedelta(days=2)
            self.assertFalse(self.conference.is_running)
            self.assertTrue(self.conference.has_ended)
            self.assertEqual(4, self.conference.days_count)

            self.assertEqual(6, self.conference.current_day)

    def test_no_start_end(self):
        with patch.object(timezone, 'now', return_value=datetime(2042, 12, 27, 12, 34, 0, 0, tzinfo=UTC)):
            self.assertFalse(self.conference.is_running)
            self.assertFalse(self.conference.has_ended)
            self.assertEqual(self.conference.days_count, 0)
            self.assertEqual(self.conference.days, [])
            self.assertEqual(self.conference.get_conference_day(timezone.now()), 0)

    def test_no_start(self):
        with patch.object(timezone, 'now', return_value=datetime(2042, 12, 27, 12, 34, 0, 0, tzinfo=UTC)):
            self.conference.end = timezone.now() + timedelta(days=2)
            self.assertFalse(self.conference.is_running)
            self.assertFalse(self.conference.has_ended)
            self.assertEqual(self.conference.days_count, 0)
            self.assertSequenceEqual(self.conference.days, [])
            self.assertEqual(self.conference.get_conference_day(timezone.now()), 0)

    def test_no_end(self):
        with patch.object(timezone, 'now', return_value=datetime(2042, 12, 27, 12, 34, 0, 0, tzinfo=UTC)):
            self.conference.start = timezone.now() - timedelta(days=2)
            self.assertTrue(self.conference.is_running)
            self.assertFalse(self.conference.has_ended)
            self.assertEqual(self.conference.days_count, 3)
            days = [
                ConferenceDay(0, self.conference.start.astimezone(self.conference.timezone), self.conference.timezone),
                ConferenceDay(1, self.conference.start.astimezone(self.conference.timezone) + timedelta(days=1), self.conference.timezone),
                ConferenceDay(2, self.conference.start.astimezone(self.conference.timezone) + timedelta(days=2), self.conference.timezone),
            ]
            # The first day starts whenever the conference starts
            days[0].start = self.conference.start
            # The last day ends at midnight not at 6 in the morning
            days[-1].end = days[-1].start.replace(hour=23, minute=59, second=59)
            self.assertSequenceEqual(self.conference.days, days)
            self.assertEqual(self.conference.get_conference_day(timezone.now()), 3)
