import re
from io import StringIO
from pathlib import Path

from babel.messages.extract import extract
from babel.messages.pofile import read_po, write_po
from jinja2.ext import babel_extract

from django.core.management.commands import makemessages


class Command(makemessages.Command):
    def handle(self, *args, **options):
        if not options['locale']:
            options['locale'] = ['de', 'en']
        options['add_location'] = 'never'
        options['no_wrap'] = True
        if not options['extensions']:
            options['extensions'] = ['html', 'htm', 'txt', 'py', 'j2']

        plainui_dir = Path(__file__).parents[2]
        self.plainui_locale_dir = plainui_dir / 'locale'
        super().handle(*args, **options)

    # intercept extracting plainui messages. Add messages extracted from jinja templates via pybabel to the pot file
    def process_locale_dir(self, locale_dir, files):
        super().process_locale_dir(locale_dir, files)

        if self.domain == 'django' and isinstance(locale_dir, (str, Path)) and self.plainui_locale_dir.samefile(locale_dir):
            pot_file = self.plainui_locale_dir / f'{self.domain}.pot'
            with pot_file.open('r') as f:
                pot_content = f.read()

            # xgettext writes this line for .pot files, pybabel does not like it and crashes
            # https://github.com/python-babel/babel/issues/665
            pot_content = pot_content.replace('"Language: \\n"\n', '')
            catalog = read_po(StringIO(pot_content), locale=None, domain=self.domain)

            for translatable_file in files:
                filename = translatable_file.path
                if not filename.endswith('.html') and not filename.endswith('.j2'):
                    continue

                with Path(filename).open('rb') as f:
                    for lineno, message, comments, context in extract(babel_extract, f):
                        catalog.add(message, None, [(filename, lineno)], auto_comments=comments, context=context)

            with pot_file.open('wb') as f:
                write_po(f, catalog, width=None, no_location=True)

    def write_po_file(self, potfile, locale):
        super().write_po_file(potfile, locale)

        basedir = Path(potfile).parent / locale / 'LC_MESSAGES'
        basedir.mkdir(exist_ok=True, parents=True)
        po_file = basedir / f'{self.domain}.po'

        with po_file.open('r', encoding='utf-8') as f:
            content = f.read()

        content = re.sub(r'"POT-Creation-Date: \d+\-\d+\-\d+ \d+:\d+\+\d+\\n"\n', '', content)

        with po_file.open(mode='w', encoding='utf-8') as f:
            f.write(content)
