__all__ = (
    'RoomView',
    'RoomsView',
)


from django.contrib import messages
from django.db.models import QuerySet
from django.http import HttpResponseRedirect
from django.utils.translation import gettext
from django.views.generic import DetailView

from core import integrations
from core.models import (
    Room,
    RoomLink,
)
from core.views.list_views import AlphabeticalPaginatorView

from plainui.views.utils import (
    ConferenceRequiredMixin,
    event_filter,
    session_get_favorite_events,
)


class RoomView(ConferenceRequiredMixin, DetailView):
    model = Room
    template_name = 'plainui/room.html.j2'
    context_object_name = 'room'

    def get_queryset(self, queryset: QuerySet | None = None) -> QuerySet[Room]:
        return Room.objects.conference_accessible(self.conf).exclude(room_type=Room.RoomType.PROJECT)

    def get(self, request, *args, **kwargs):
        self.room = self.object = self.get_object()

        try:
            if self.room.room_type == Room.RoomType.BIGBLUEBUTTON:
                if self.request.user.is_authenticated:
                    return HttpResponseRedirect(integrations.BigBlueButton.join_room(self.room, self.request.user, not self.request.user.show_name))

                messages.error(request, gettext('RoomView--login_required'))

        except integrations.IntegrationError as e:
            messages.error(request, str(e))

        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf

        linking_links = []
        media_link = None
        for link in self.room.links.all():
            if link.link_type == RoomLink.LinkType.MEDIA_CCC_DE:
                media_link = link.link
            elif link.link_type in [RoomLink.LinkType.WEBSITE, RoomLink.LinkType.CHAT, RoomLink.LinkType.JITSI, RoomLink.LinkType.PAD, RoomLink.LinkType.NAV]:
                linking_links.append(link)

        context['voc_stream'] = media_link
        context['links'] = linking_links
        context['events'] = event_filter(self.request.user, self.conf, room=self.room)
        context['my_favorite_events'] = session_get_favorite_events(self.request.session, self.request.user)

        context['report_info'] = {'lookup_key': self.room.slug}

        return context


class RoomsView(ConferenceRequiredMixin, AlphabeticalPaginatorView):
    model = Room
    template_name = 'plainui/rooms.html.j2'
    context_object_name = 'rooms'
    filters = {
        'name': ('name__icontains', 'str'),
    }
    ordering = ['name']

    def get_queryset(self, queryset: QuerySet | None = None) -> QuerySet[Room]:
        return Room.objects.conference_accessible(self.conf).exclude(room_type=Room.RoomType.PROJECT)

    def get_context_data(self, **kwargs):
        return {
            **super().get_context_data(**kwargs),
            'conf': self.conf,
            'report_info': {'enabled': True},
        }
