import threading
from datetime import UTC, datetime, timedelta
from typing import ClassVar

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import mixins as auth_mixins
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db.models import Case, F, Prefetch, Q, When
from django.forms import BooleanField
from django.http import HttpResponse
from django.shortcuts import redirect
from django.urls import reverse
from django.utils import timezone
from django.utils.timezone import localtime
from django.utils.translation import gettext

from core.models import (
    Assembly,
    Conference,
    Event,
    EventParticipant,
    PlatformUser,
    Project,
    Room,
    TagItem,
)


class ConferenceRequiredMixin(auth_mixins.AccessMixin):
    require_login = True
    """ This view is restricted to logged-in users if the conference is non-public """
    require_user = False
    """ This view requires a logged-in user, regardless of the conference being public or non-public """
    require_conference_member = True
    """ This view requires the user having a valid ticket. Implies `require_login`. """
    only_published = True
    """ This view should only be shown when the conference is published."""

    # cache for the conference used in this plainui instance
    _conf: ClassVar[Conference | None] = None
    _conf_timeout: ClassVar[datetime] = datetime.min.replace(tzinfo=UTC)
    _conf_lock: ClassVar[threading.Lock] = threading.Lock()

    # instance variables
    conf: Conference
    now: datetime

    # flag for internal usage (in the tests): will be set to True there to test the mixin behaviour without the need
    # to call the view(s) - accomplished by not calling super() in dispatch() and returning an empty response instead
    _test_cork = False

    def get_login_url(self):
        return reverse('plainui:login')

    def has_ticket(self, request) -> bool:
        if not request.user.is_authenticated:
            return False

        session_conferences = request.session.get('conferences', [])
        if str(self.conf.id) in session_conferences:
            return True

        if self.conf.users.filter(user=request.user, has_ticket=True).exists():
            session_conferences.append(str(self.conf.id))
            request.session['conferences'] = session_conferences
            return True
        else:
            return False

    def dispatch(self, request, *args, **kwargs):
        conf = ConferenceRequiredMixin._conf
        self.now = now = timezone.now()
        if conf is None or self._conf_timeout < now:
            if self._conf_lock.acquire(blocking=conf is None):
                try:
                    conferences = Conference.objects.all()
                    try:
                        if settings.SELECTED_CONFERENCE_ID:
                            conferences = conferences.filter(id=settings.SELECTED_CONFERENCE_ID)
                        ConferenceRequiredMixin._conf = conf = conferences.get()
                        self._conf_timeout = now + timedelta(300)
                    except Conference.MultipleObjectsReturned:
                        raise ImproperlyConfigured('Multiple Conferences found, set SELECTED_CONFERENCE_ID in settings!')
                    except Conference.DoesNotExist:
                        if Conference.objects.exists():
                            raise ImproperlyConfigured('SELECTED_CONFERENCE_ID from settings not found in database!')
                        else:
                            raise ImproperlyConfigured('No conference found in database!')
                    except ValidationError:
                        raise ImproperlyConfigured('Invalid value for SELECTED_CONFERENCE_ID in settings!')
                finally:
                    self._conf_lock.release()

        self.conf = conf

        if self.only_published and not conf.is_published:
            return redirect(reverse('plainui:under_construction'))

        if (
            (self.require_login and conf.require_login) or (conf.require_ticket and self.require_conference_member) or self.require_user
        ) and not request.user.is_authenticated:
            return self.handle_no_permission()

        if conf.require_ticket and self.require_conference_member and not self.has_ticket(request):
            messages.error(request, gettext('Please activate your Ticket to access this conference!'))
            return redirect(reverse('plainui:redeem_token'))

        if self._test_cork:
            return HttpResponse()

        return super().dispatch(request, *args, **kwargs)


def organize_events_for_calendar(conf, events):
    # group events by their rooms, insert spacers between events if necessary
    events_by_room = {}
    room_end = {}
    first_start = None
    for event in events:
        if first_start is None:
            first_start = event.schedule_start
            calendar_start = localtime(first_start.replace(minute=0, second=0, microsecond=0))
            if localtime(first_start) != calendar_start:
                first_start = first_start - (localtime(first_start) - calendar_start)

        if event.room_id not in events_by_room:
            events_by_room[event.room_id] = []
            room_end[event.room_id] = first_start

        free_time = event.schedule_start - room_end[event.room_id]
        if free_time != timedelta(0):
            events_by_room[event.room_id].append({'type': 'space', 'minutes': free_time.total_seconds() / 60})
        events_by_room[event.room_id].append({'type': 'event', 'event': event, 'minutes': event.schedule_duration.total_seconds() / 60})
        room_end[event.room_id] = event.schedule_end

    if not events_by_room:
        return None

    rooms = (
        Room.objects.conference_accessible(conf)
        .filter(pk__in=events_by_room.keys())
        .order_by('official_room_order', F('capacity').desc(nulls_last=True), 'name')
    )
    rooms_with_events = [(room, events_by_room[room.id]) for room in rooms]

    # calendar starts at first scheduled event rounded down to the full hour
    calendar_start = min(
        room_events[0]['event'].schedule_start if room_events[0]['type'] == 'event' else room_events[1]['event'].schedule_start
        for (_, room_events) in rooms_with_events
    )
    calendar_start = localtime(calendar_start.replace(minute=0, second=0, microsecond=0))

    # calendar ends at the end of the last scheduled event rounded up to the full hour
    calendar_end = max(room_events[-1]['event'].schedule_end for (_, room_events) in rooms_with_events)
    calendar_end = (calendar_end + timedelta(0, 3599)).replace(minute=0, second=0, microsecond=0)
    calendar_end = localtime(calendar_end)

    # put timestamps from calendar_start to calendar_end in 30 minute steps into calendar_time_steps
    calendar_time_steps = []
    calendar_step_minutes = 30
    current_tick = calendar_start
    current_idx = None
    while current_tick < calendar_end:
        day_idx = conf.get_conference_day(current_tick)
        calendar_time_steps.append({'day': day_idx, 'new_day': day_idx != current_idx, 'ts': current_tick})
        current_idx = day_idx
        current_tick += timedelta(0, 60 * calendar_step_minutes)

    return {
        'rooms_with_events': rooms_with_events,
        'calendar_start': calendar_start,
        'calendar_end': calendar_end,
        'calendar_time_steps': calendar_time_steps,
        'calendar_step_minutes': calendar_step_minutes,
    }


UPCOMING_WINDOW = timedelta(minutes=30)


def event_filter(
    user: PlatformUser,
    conf,
    day=None,
    kinds=None,
    assembly=None,
    track=None,
    room=None,
    user_schedule_only=False,
    upcoming=False,
    calendar_mode=True,
    public_fahrplan=None,
    is_recorded: bool | None = None,
):
    min_date, max_date = conf.start, conf.end
    if min_date is None or max_date is None:
        return Event.objects.none()

    filters = {}
    if day is not None:
        filters['schedule_start__gte'] = min_date + timedelta(day)
        filters['schedule_start__lt'] = min_date + timedelta(day + 1)
    if kinds is not None:
        filters['kind__in'] = kinds
    if assembly:
        filters['assembly'] = assembly
    if track:
        filters['track'] = track
    if room:
        filters['room'] = room
    if public_fahrplan is not None:
        filters['room__is_public_fahrplan'] = public_fahrplan
    if user_schedule_only and user.is_authenticated:
        filters['favorite_of'] = user

    events = Event.objects.conference_accessible(conf)
    events = events.filter(schedule_start__isnull=False, schedule_end__isnull=False)
    now = timezone.now()
    events = events.annotate(
        upcoming=Case(
            When(Q(schedule_start__lt=now, schedule_end__gte=now) | Q(schedule_start__gte=now, schedule_start__lt=now + UPCOMING_WINDOW)),
            default=False,
            output_field=BooleanField(),
        )
    )
    if upcoming:
        events = events.filter(upcoming=True)
    if calendar_mode:
        events = events.filter(room__isnull=False)
    res = events.filter(schedule_duration__isnull=False, **filters).order_by('schedule_start', 'schedule_end')

    # filter on the requested recording state
    if is_recorded is True:
        # don't include all events explicitly labeled "don't record"
        res = res.exclude(recording=Event.Recording.NO)

        res = res.filter(
            # ... all events of unknown state in rooms with "record by default"
            Q(recording=Event.Recording.UNKNOWN, room__isnull=False, room__recording_state__in=[Room.RecordingState.RECORD_BY_DEFAULT])
            # ... all events with state "yes, record" which either ...
            | (
                Q(recording=Event.Recording.YES)
                & (
                    # don't have a room
                    Q(room__isnull=True)
                    # have a room which does allow recording
                    | Q(room__isnull=False, room__recording_state__in=[Room.RecordingState.NOT_RECORDED_BY_DEFAULT, Room.RecordingState.RECORD_BY_DEFAULT])
                )
            )
        )

    elif is_recorded is False:
        # kick all events which don't have a room and don't explicitly specify not to be recorded
        res = res.exclude(room__isnull=True, recording__in=[Event.Recording.UNKNOWN, Event.Recording.YES])

        # kick all events which ...
        res = res.exclude(
            # shall explicitly be recorded and ...
            recording=Event.Recording.YES,
            # have a room which permits recording (if an event may not be recorded because of this we can't filter it even though it requests to be recorded)
            room__isnull=False,
            room__recording_state__in=[Room.RecordingState.RECORD_BY_DEFAULT, Room.RecordingState.NOT_RECORDED_BY_DEFAULT],
        )

        # kick all events which default to their room's setting which would allow recording
        res = res.exclude(recording=Event.Recording.UNKNOWN, room__isnull=False, room__recording_state__in=[Room.RecordingState.RECORD_BY_DEFAULT])

    res = res.annotate(track_name=F('track__name'))
    speakers = EventParticipant.objects.filter(is_public=True, role=EventParticipant.Role.SPEAKER).order_by('order', 'participant__display_name')
    speakers = speakers.annotate(speaker_name=F('participant__display_name'))

    return res.prefetch_related(
        Prefetch('participants', queryset=speakers, to_attr='speakers'),
        Prefetch('tags', to_attr='prefetched_tags', queryset=TagItem.objects.select_related('tag').filter(tag__is_public=True)),
    ).select_related('room', 'assembly', 'conference')


_SESSION_TYPE_MAP = {
    'fav_a': Assembly,
    'fav_e': Event,
    'fav_p': Project,
}


def _session_refresh_favorite(session, user, object_type: str) -> list[str]:
    if not user.is_authenticated:
        return []

    favorites = [str(id_) for id_ in _SESSION_TYPE_MAP[object_type].objects.filter(favorite_of=user).values_list('id', flat=True)]
    session[object_type] = favorites
    return favorites


def session_get_favorite(session, user, object_type: str) -> list[str]:
    if not user.is_authenticated:
        return []

    favorites = session.get(object_type, None)
    if favorites is None:
        return _session_refresh_favorite(session, user, object_type)
    return favorites


def session_get_favorite_assemblies(session, user) -> list[str]:
    return session_get_favorite(session, user, 'fav_a')


def _session_refresh_favorite_assemblies(session, user) -> list[str]:
    return _session_refresh_favorite(session, user, 'fav_a')


def session_get_favorite_events(session, user) -> list[str]:
    return session_get_favorite(session, user, 'fav_e')


def _session_refresh_favorite_events(session, user) -> list[str]:
    return _session_refresh_favorite(session, user, 'fav_e')


def session_get_favorite_projects(session, user) -> list[str]:
    return session_get_favorite(session, user, 'fav_p')


def _session_refresh_favorite_projects(session, user) -> list[str]:
    return _session_refresh_favorite(session, user, 'fav_p')
